"""Module describing the planemo ``tool_init`` command."""

from typing import (
    Any,
    Dict,
)

import click

from planemo import (
    io,
    options,
    tool_builder,
)
from planemo.cli import command_function


@click.command("tool_init")
@options.tool_init_id_option()
@options.force_option(what="tool")
@options.tool_init_tool_option()
@options.tool_init_name_option()
@options.tool_init_version_option()
@options.tool_init_description_option()
@options.tool_init_command_option()
@options.tool_init_example_command_option()
@options.tool_init_example_input_option()
@options.tool_init_example_output_option()
@options.tool_init_named_output_option()
@options.tool_init_input_option()
@options.tool_init_output_option()
@options.tool_init_help_text_option()
@options.tool_init_help_from_command_option()
@options.tool_init_doi_option()
@options.tool_init_cite_url_option()
@options.tool_init_test_case_option()
@options.tool_init_macros_option()
@options.tool_init_version_command_option()
@options.tool_init_requirement_option()
@options.tool_init_container_option()
@options.build_cwl_option()
@command_function
def cli(ctx, **kwds):
    """Generate tool outline from given arguments."""
    invalid = _validate_kwds(kwds)
    if invalid:
        ctx.exit(invalid)
    tool_description = tool_builder.build(**kwds)
    tool_builder.write_tool_description(ctx, tool_description, **kwds)


def _validate_kwds(kwds: Dict[str, Any]) -> int:
    def not_exclusive(x, y):
        if kwds.get(x) and kwds.get(y):
            io.error(f"Can only specify one of --{x} and --{y}")
            return True

    def not_specifing_dependent_option(x, y):
        if kwds.get(x) and not kwds.get(y):
            template = "Can only use the --%s option if also specifying --%s"
            message = template % (x, y)
            io.error(message)
            return True

    if not_exclusive("help_text", "help_from_command"):
        return 1
    if not_exclusive("command", "example_command"):
        return 1
    if not_specifing_dependent_option("example_input", "example_command"):
        return 1
    if not_specifing_dependent_option("example_output", "example_command"):
        return 1
    if not_specifing_dependent_option("test_case", "example_command"):
        return 1
    return 0



# file auto generated with scripts/tool_index_to_id_map.py
DISTRO_TOOLS_ID_TO_PATH = {
    "ucsc_table_direct1": "data_source/ucsc_tablebrowser.xml",
    "MAF_To_Fasta1": "maf/maf_to_fasta.xml",
    "ratmine": "data_source/ratmine.xml",
    "mousemine": "data_source/mousemine.xml",
    "__SORTLIST__": "${model_tools_path}/sort_collection_list.xml",
    "__UNZIP_COLLECTION__": "${model_tools_path}/unzip_collection.xml",
    "cbi_rice_mart": "data_source/cbi_rice_mart.xml",
    "ucsc_table_direct_archaea1": "data_source/ucsc_tablebrowser_archaea.xml",
    "wig_to_bigWig": "filters/wig_to_bigwig.xml",
    "ebi_sra_main": "data_source/ebi_sra.xml",
    "metabolicmine": "data_source/metabolicmine.xml",
    "Extract_features1": "filters/gff/extract_GFF_Features.xml",
    "upload1": "data_source/upload.xml",
    "wc_gnu": "filters/wc_gnu.xml",
    "join1": "filters/joiner.xml",
    "random_lines1": "filters/randomlines.xml",
    "modENCODEfly": "data_source/fly_modencode.xml",
    "gff_filter_by_attribute": "filters/gff/gff_filter_by_attribute.xml",
    "gtf2bedgraph": "filters/gtf2bedgraph.xml",
    "ChangeCase": "filters/changeCase.xml",
    "__FLATTEN__": "${model_tools_path}/flatten_collection.xml",
    "wiggle2simple1": "filters/wiggle_to_simple.xml",
    "GeneBed_Maf_Fasta2": "maf/genebed_maf_to_fasta.xml",
    "trimmer": "filters/trimmer.xml",
    "createInterval": "filters/CreateInterval.xml",
    "gff_filter_by_feature_count": "filters/gff/gff_filter_by_feature_count.xml",
    "Interval2Maf1": "maf/interval2maf.xml",
    "genomespace_exporter": "genomespace/genomespace_exporter.xml",
    "Show tail1": "filters/tailWrapper.xml",
    "barchart_gnuplot": "plotting/bar_chart.xml",
    "microbial_import1": "data_source/microbial_import.xml",
    "axt_to_concat_fasta": "filters/axt_to_concat_fasta.xml",
    "__FILTER_FROM_FILE__": "${model_tools_path}/filter_from_file.xml",
    "Interval2Maf_pairwise1": "maf/interval2maf_pairwise.xml",
    "Show beginning1": "filters/headWrapper.xml",
    "axt_to_lav_1": "filters/axt_to_lav.xml",
    "modmine": "data_source/modmine.xml",
    "__EXPORT_HISTORY__": "${model_tools_path}/imp_exp/exp_history_to_archive.xml",
    "MAF_Thread_For_Species1": "maf/maf_thread_for_species.xml",
    "vcf_to_maf_customtrack1": "maf/vcf_to_maf_customtrack.xml",
    "MAF_To_BED1": "maf/maf_to_bed.xml",
    "__DATA_FETCH__": "${model_tools_path}/data_fetch.xml",
    "__IMPORT_HISTORY__": "${model_tools_path}/imp_exp/imp_history_from_archive.xml",
    "biomart": "data_source/biomart.xml",
    "Sff_extractor": "filters/sff_extractor.xml",
    "secure_hash_message_digest": "filters/secure_hash_message_digest.xml",
    "MAF_Reverse_Complement_1": "maf/maf_reverse_complement.xml",
    "mergeCols1": "filters/mergeCols.xml",
    "gff2bed1": "filters/gff2bed.xml",
    "Grouping1": "stats/grouping.xml",
    "maf_limit_size1": "maf/maf_limit_size.xml",
    "sort1": "filters/sorter.xml",
    "Convert characters1": "filters/convert_characters.xml",
    "MAF_To_Interval1": "maf/maf_to_interval.xml",
    "MAF_filter": "maf/maf_filter.xml",
    "MAF_split_blocks_by_species1": "maf/maf_split_by_species.xml",
    "genomespace_importer": "genomespace/genomespace_importer.xml",
    "gene2exon1": "filters/ucsc_gene_bed_to_exon_bed.xml",
    "Cut1": "filters/cutWrapper.xml",
    "Count1": "filters/uniq.xml",
    "MAF_Limit_To_Species1": "maf/maf_limit_to_species.xml",
    "ucsc_table_direct_test1": "data_source/ucsc_tablebrowser_test.xml",
    "wormbase": "data_source/wormbase.xml",
    "maf_stats1": "maf/maf_stats.xml",
    "zebrafishmine": "data_source/zebrafishmine.xml",
    "Paste1": "filters/pasteWrapper.xml",
    "Interval_Maf_Merged_Fasta2": "maf/interval_maf_to_merged_fasta.xml",
    "modENCODEworm": "data_source/worm_modencode.xml",
    "gtf_filter_by_attribute_values_list": "filters/gff/gtf_filter_by_attribute_values_list.xml",
    "Summary_Statistics1": "stats/gsummary.xml",
    "qual_stats_boxplot": "plotting/boxplot.xml",
    "cat1": ["filters/catWrapper.xml", "filters/catWrapper2.xml"],
    "maf_by_block_number1": "maf/maf_by_block_number.xml",
    "Grep1": "filters/grep.xml",
    "eupathdb": "data_source/eupathdb.xml",
    "__RELABEL_FROM_FILE__": "${model_tools_path}/relabel_from_file.xml",
    "__ZIP_COLLECTION__": "${model_tools_path}/zip_collection.xml",
    "lav_to_bed1": "filters/lav_to_bed.xml",
    "comp1": "filters/compare.xml",
    "bed_to_bigBed": "filters/bed_to_bigbed.xml",
    "liftOver1": "extract/liftOver_wrapper.xml",
    "bed2gff1": "filters/bed2gff.xml",
    "__MERGE_COLLECTION__": "${model_tools_path}/merge_collection.xml",
    "gramenemart": "data_source/gramene_mart.xml",
    "hbvar": "data_source/hbvar.xml",
    "yeastmine": "data_source/yeastmine.xml",
    "Filter1": "stats/filtering.xml",
    "flymine": "data_source/flymine.xml",
    "wormbase_test": "data_source/wormbase_test.xml",
    "__FILTER_FAILED_DATASETS__": "${model_tools_path}/filter_failed_collection.xml",
    "Extract genomic DNA 1": "extract/extract_genomic_dna.xml",
    "Remove beginning1": "filters/remove_beginning.xml",
    "flymine_test": "data_source/flymine_test.xml",
    "axt_to_fasta": "filters/axt_to_fasta.xml",
    "addValue": "filters/fixedValueColumn.xml",
    "__TAG_FROM_FILE__": "${model_tools_path}/tag_collection_from_file.xml",
    "param_value_from_file": "expression_tools/parse_values_from_file.xml",
    "__APPLY_RULES__": "${model_tools_path}/apply_rules.xml",
}



"""Module describing the planemo ``rerun`` command."""

import click

from planemo import options
from planemo.cli import command_function
from planemo.engine import engine_context
from planemo.io import (
    error,
    info,
)
from planemo.runnable import Rerunnable


@click.command("rerun")
@options.profile_option()
@options.galaxy_url_option()
@options.galaxy_user_key_option()
@click.option(
    "--invocation",
    "rerunnable_type",
    flag_value="invocation",
    help=("Rerun failed jobs associated by one or more invocation IDs."),
)
@click.option(
    "--history",
    "rerunnable_type",
    flag_value="history",
    help=("Rerun failed jobs associated by one or more history IDs."),
)
@click.option(
    "--job", "rerunnable_type", flag_value="job", help=("Rerun failed jobs specified by one or more job IDs.")
)
@click.argument(
    "rerunnable_ids",
    metavar="RERUNNABLE_IDS",
    type=str,
    nargs=-1,
)
@command_function
def cli(ctx, rerunnable_ids, **kwds):
    """Planemo command for rerunning and remapping failed jobs on an external Galaxy server.
    Supply a list of history, invocation or job IDs, identifying the ID type using the
    --invocation, --history or --job flag, and all associated failed jobs will be rerun.

    Please note: attempting to rerun non-remappable jobs will result in an exit code of 1. As
    jobs cannot be remapped more than once, running this command two or more times with the same
    history or job IDs will therefore return an exit code of 1. If avoiding this is important,
    you should specify the invocation ID instead if possible.

    \b
        % planemo rerun --invocation / --history / --job RERUNNABLE_IDS
    """
    # Possible TODO: allow collection IDs to be specified as well
    if not kwds.get("rerunnable_type"):
        error("Please specify the type (invocation, history or job) of the IDs which should be rerun.")
        ctx.exit(1)
    kwds["engine"] = "external_galaxy"
    rerun_successful = True
    with engine_context(ctx, **kwds) as engine:
        for rerunnable_id in rerunnable_ids:
            rerunnable = Rerunnable(rerunnable_id, kwds["rerunnable_type"], kwds["galaxy_url"])
            rerun_result = engine.rerun(ctx, rerunnable, **kwds)
            if not rerun_result.was_successful:
                rerun_successful = False

    if rerun_successful:
        info("All requested jobs were rerun successfully.")
        ctx.exit(0)
    else:
        error("Some of the requested jobs could not be rerun.")
        ctx.exit(1)



""" Utilities for using virtualenv as library and planemo command.
"""

import os
import sys
from typing import Optional

from galaxy.util.commands import which

DEFAULT_PYTHON_VERSION = os.environ.get("PLANEMO_DEFAULT_PYTHON_VERSION", "3")


def create_command(virtualenv_path: str, galaxy_python_version: Optional[str] = None) -> str:
    """If virtualenv is on Planemo's path use it, otherwise use the planemo
    subcommand virtualenv to create the virtualenv.
    """
    # Create a virtualenv with the selected python version.
    if galaxy_python_version is None:
        galaxy_python_version = DEFAULT_PYTHON_VERSION
    python = which("python%s" % galaxy_python_version)
    if python:
        python = os.path.abspath(python)
    else:
        python = sys.executable or "python"
    virtualenv_on_path = which("virtualenv")
    if virtualenv_on_path:
        command = [virtualenv_on_path, virtualenv_path, "-p", python]
    else:
        command = [python, "-m", "venv", virtualenv_path]
    return " ".join(command)



"""Module describing the planemo ``upload_data`` command."""

import click

from planemo import options
from planemo.cli import command_function
from planemo.engine import engine_context
from planemo.galaxy.activity import stage_in
from planemo.galaxy.workflows import rewrite_job_file
from planemo.io import info
from planemo.runnable_resolve import for_runnable_identifier


@click.command("upload_data")
@options.required_runnable_arg()
@options.required_job_arg()
@options.required_new_job_arg()
@options.galaxy_run_options()
@options.galaxy_config_options()
@options.run_history_tags_option()
@command_function
def cli(ctx, runnable_identifier, job_path, new_job_path, **kwds):
    """Planemo command for uploading data to an external Galaxy server.

    \b
        % planemo upload_data wf.ga wf-job.yml new-wf-job.yml --profile profile

    Running this subcommand requires a workflow file or identifier
    and a job file, just as for ``planemo run``. In addition, a third
    argument is required, for the location of a new job file which
    Planemo will write. The data will be uploaded to the specified
    external Galaxy server and the job file will be recreated at the
    specified location, with all instances of ``path`` or ``location``
    for input datasets and collections replaced by ``galaxy_id``. The
    new job file can then be used to run the workflow separately from
    the already completed data upload.
    """
    kwds["engine"] = "external_galaxy"  # to only upload data to a local Galaxy does not make sense
    runnable = for_runnable_identifier(ctx, runnable_identifier, kwds)
    with engine_context(ctx, **kwds) as engine:
        with engine.ensure_runnables_served([runnable]) as config:
            job, _, _ = stage_in(ctx, runnable, config, job_path, **kwds)

    rewrite_job_file(job_path, new_job_path, job)
    info(f"Files uploaded and new job file written to {new_job_path}")

    return 0



"""Module describing the planemo ``delete_alias`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.galaxy import profiles
from planemo.io import (
    error,
    info,
)

try:
    from tabulate import tabulate
except ImportError:
    tabulate = None  # type: ignore


@click.command("delete_alias")
@options.alias_option(required=True)
@options.profile_option(required=True)
@command_function
def cli(ctx, alias, profile, **kwds):
    """
    List aliases for a path or a workflow or dataset ID. Aliases are associated with a particular planemo profile.
    """
    info("Looking for profiles...")
    exit_code = profiles.delete_alias(ctx, alias, profile)
    if exit_code == 0:
        info(f"Alias {alias} was successfully deleted from profile {profile}")
    else:
        error(f"Alias {alias} does not exist, so was not deleted from profile {profile}")

    ctx.exit(exit_code)
    return



"""Module contains code for the Topic class, dealing with the creation of a training topic."""

import collections
import os

from planemo import templates
from .utils import (
    load_yaml,
    Requirement,
    save_to_yaml,
)

INDEX_FILE_TEMPLATE = """---
layout: topic
topic_name: {{ topic }}
---
"""

README_FILE_TEMPLATE = """
{{ topic }}
==========

Please refer to the [CONTRIBUTING.md](../../CONTRIBUTING.md) before adding or updating any material
"""


DOCKER_FILE_TEMPLATE = """
# Galaxy - {{ topic_title }}
#
# to build the docker image, go to root of training repo and
#    docker build -t {{ topic_name }} -f topics/{{ topic_name }}/docker/Dockerfile .
#
# to run image:
#    docker run -p "8080:80" -t {{ topic_name }}
#    use -d to automatically dowload the datalibraries in the container

FROM quay.io/bgruening/galaxy:20.05

MAINTAINER Galaxy Training Material

ENV GALAXY_CONFIG_BRAND "GTN: {{ topic_title }}"

# copy the tutorials directory for your topic
ADD topics/{{ topic_name }}/tutorials/ /tutorials/

# install everything for tutorials
ADD bin/docker-install-tutorials.sh /setup-tutorials.sh
ADD bin/mergeyaml.py /mergeyaml.py
ADD bin/data_libarary_download.sh /data_libarary_download.sh
RUN /setup-tutorials.sh

ENTRYPOINT ["/data_libarary_download.sh"]
"""


INTRO_SLIDES_FILE_TEMPLATE = """---
layout: introduction_slides
logo: "GTN"

title: {{ title }}
type: {{ type }}
contributors:
- contributor
---

### How to fill the slide decks?

Please follow our
[tutorial to learn how to fill the slides]({{ '{{' }} site.baseurl {{ '}}' }}/topics/contributing/tutorials/create-new-tutorial-slides/slides.html)
"""


class Topic:
    """Class to describe a training topic."""

    def __init__(self, name="new_topic", target="use", title="The new topic", summary="Summary", parent_dir="topics"):
        """Init a topic instance."""
        self.name = name
        self.type = target
        self.title = title
        self.summary = summary
        self.docker_image = ""
        self.maintainers = ["maintainers"]
        self.parent_dir = parent_dir
        self.set_default_requirement()
        self.set_paths()

    def init_from_kwds(self, kwds):
        """Init a topic instance from a kwds dictionary."""
        self.name = kwds["topic_name"]
        self.type = kwds["topic_target"]
        self.title = kwds["topic_title"]
        self.summary = kwds["topic_summary"]
        self.set_default_requirement()
        self.set_paths()

    def init_from_metadata(self):
        """Init a topic instance from the metadata file."""
        metadata = load_yaml(self.metadata_fp)
        self.name = metadata["name"]
        self.type = metadata["type"]
        self.title = metadata["title"]
        self.summary = metadata["summary"]
        self.requirements = []
        if "requirements" in metadata:
            for r in metadata["requirements"]:
                req = Requirement()
                req.init_from_dict(r)
                self.requirements.append(req)
        if "docker_image" in metadata:
            self.docker_image = metadata["docker_image"]
        self.maintainers = metadata["maintainers"]
        self.set_paths()

    # GETTERS
    def get_requirements(self):
        """Get the requirements as a list of ordered dictionaries."""
        reqs = []
        for req in self.requirements:
            reqs.append(req.export_to_ordered_dict())
        return reqs

    def export_metadata_to_ordered_dict(self):
        """Export the topic metadata into an ordered dictionary."""
        metadata = collections.OrderedDict()
        metadata["name"] = self.name
        metadata["type"] = self.type
        metadata["title"] = self.title
        metadata["summary"] = self.summary
        metadata["requirements"] = self.get_requirements()
        metadata["docker_image"] = self.docker_image
        metadata["maintainers"] = self.maintainers
        return metadata

    # SETTERS
    def set_default_requirement(self):
        """Set default requirement: Galaxy introduction."""
        self.requirements = []
        if self.type == "use":
            self.requirements.append(Requirement())

    def set_paths(self):
        """Set the paths to folder and files."""
        self.dir = os.path.join(self.parent_dir, self.name)
        self.img_folder = os.path.join(self.dir, "images")
        self.tuto_folder = os.path.join(self.dir, "tutorials")
        self.index_fp = os.path.join(self.dir, "index.md")
        self.readme_fp = os.path.join(self.dir, "README.md")
        self.metadata_fp = os.path.join(self.dir, "metadata.yaml")
        self.docker_folder = os.path.join(self.dir, "docker")
        self.dockerfile_fp = os.path.join(self.docker_folder, "Dockerfile")
        self.slides_folder = os.path.join(self.dir, "slides")

    # TESTS
    def exists(self):
        """Test if the topic exists."""
        return os.path.isdir(self.dir)

    # OTHER METHODS
    def create_topic_structure(self):
        """Create the skeleton of a new topic.

        1. create the folder and its structure
        2. update the index.md to match your topic's name
        3. fill the metadata
        4. add a symbolic link to the metadata.yaml from the metadata folder
        """
        # create the folder and its structure
        os.makedirs(self.dir)
        self.img_folder = os.path.join(self.dir, "images")
        os.makedirs(self.img_folder)
        self.tuto_folder = os.path.join(self.dir, "tutorials")
        os.makedirs(self.tuto_folder)

        # create the index.md and add the topic name
        self.index_fp = os.path.join(self.dir, "index.md")
        with open(self.index_fp, "w") as index_f:
            index_f.write(templates.render(INDEX_FILE_TEMPLATE, **{"topic": self.name}))

        # create the README file
        self.readme_fp = os.path.join(self.dir, "README.md")
        with open(self.readme_fp, "w") as readme_f:
            readme_f.write(templates.render(README_FILE_TEMPLATE, **{"topic": self.title}))

        # create the metadata file
        self.metadata_fp = os.path.join(self.dir, "metadata.yaml")
        save_to_yaml(self.export_metadata_to_ordered_dict(), self.metadata_fp)

        # create Dockerfile
        self.docker_folder = os.path.join(self.dir, "docker")
        os.makedirs(self.docker_folder)
        self.dockerfile_fp = os.path.join(self.docker_folder, "Dockerfile")
        with open(self.dockerfile_fp, "w") as dockerfile:
            dockerfile.write(
                templates.render(DOCKER_FILE_TEMPLATE, **{"topic_name": self.name, "topic_title": self.title})
            )

        # create empty introduction slides
        self.slides_folder = os.path.join(self.dir, "slides")
        os.makedirs(self.slides_folder)
        self.intro_slide_fp = os.path.join(self.slides_folder, "introduction.html")
        with open(self.intro_slide_fp, "w") as intro_slide_f:
            intro_slide_f.write(
                templates.render(
                    INTRO_SLIDES_FILE_TEMPLATE, **{"title": "Introduction to %s" % self.title, "type": "introduction"}
                )
            )

        # add a symbolic link to the metadata.yaml
        metadata_dir = "metadata"
        if not os.path.isdir(metadata_dir):
            os.makedirs(metadata_dir)
        os.chdir(metadata_dir)
        os.symlink(os.path.join("..", self.metadata_fp), "%s.yaml" % self.name)
        os.chdir("..")



"""Module describing the planemo ``create_alias`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.galaxy import profiles
from planemo.io import info

try:
    import namesgenerator
except ImportError:
    namesgenerator = None


@click.command("create_alias")
@click.argument(
    "obj",
    metavar="OBJ",
    type=click.STRING,
)
@options.alias_option()
@options.profile_option(required=True)
@command_function
def cli(ctx, alias, obj, profile, **kwds):
    """
    Add an alias for a path or a workflow or dataset ID. Aliases are associated with a particular planemo profile.
    """
    if not alias:
        if not namesgenerator:
            raise ImportError(
                "Random generation of aliases requires installation of the namesgenerator package."
                "Either install this, or specify the alias name with --alias."
            )
        alias = namesgenerator.get_random_name()

    exit_code = profiles.create_alias(ctx, alias, obj, profile)
    info(f"Alias {alias} created.")
    ctx.exit(exit_code)
    return



"""This module contains :func:`build` to build tool descriptions.

This class is used by the `tool_init` command and can be used to build
Galaxy and CWL tool descriptions.
"""

import os
import re
import shlex
import shutil
import subprocess
from collections import namedtuple

from planemo import (
    io,
    templates,
)

REUSING_MACROS_MESSAGE = (
    "Macros file macros.xml already exists, assuming " " it has relevant planemo-generated definitions."
)
DEFAULT_CWL_VERSION = "v1.0"


TOOL_TEMPLATE = """<tool id="{{id}}" name="{{name}}" version="{{version}}+galaxy0" python_template_version="3.5" profile="21.05">
{%- if description %}
    <description>{{ description }}</description>
{%- endif %}
{%- if macros %}
    <macros>
        <import>macros.xml</import>
    </macros>
    <expand macro="requirements" />
{%- if version_command %}
    <expand macro="version_command" />
{%- endif %}
{%- else %}
    <requirements>
{%- for requirement in requirements %}
        {{ requirement }}
{%- endfor %}
{%- for container in containers %}
        {{ container }}
{%- endfor %}
    </requirements>
{%- if version_command %}
    <version_command>{{ version_command }}</version_command>
{%- endif %}
{%- endif %}
    <command detect_errors="exit_code"><![CDATA[
{%- if command %}
        {{ command }}
{%- else %}
        TODO: Fill in command template.
{%- endif %}
    ]]></command>
    <inputs>
{%- for input in inputs %}
        {{ input }}
{%- endfor %}
    </inputs>
    <outputs>
{%- for output in outputs %}
        {{ output }}
{%- endfor %}
    </outputs>
{%- if tests %}
    <tests>
{%- for test in tests %}
        <test>
{%- for param in test.params %}
            <param name="{{ param[0]}}" value="{{ param[1] }}"/>
{%- endfor %}
{%- for output in test.outputs %}
            <output name="{{ output[0] }}" file="{{ output[1] }}"/>
{%- endfor %}
        </test>
{%- endfor %}
    </tests>
{%- endif %}
    <help><![CDATA[
{%- if help %}
        {{ help }}
{%- else %}
        TODO: Fill in help.
{%- endif %}
    ]]></help>
{%- if macros %}
    <expand macro="citations" />
{%- else %}
{%- if doi or bibtex_citations %}
    <citations>
{%- for single_doi in doi %}
        <citation type="doi">{{ single_doi }}</citation>
{%- endfor %}
{%- for bibtex_citation in bibtex_citations %}
        <citation type="bibtex">{{ bibtex_citation }}</citation>
{%- endfor %}
    </citations>
{%- endif %}
{%- endif %}
</tool>
"""

MACROS_TEMPLATE = """<macros>
    <xml name="requirements">
        <requirements>
{%- for requirement in requirements %}
        {{ requirement }}
{%- endfor %}
            <yield/>
{%- for container in containers %}
        {{ container }}
{%- endfor %}
        </requirements>
    </xml>
    <xml name="citations">
        <citations>
{%- for single_doi in doi %}
            <citation type="doi">{{ single_doi }}</citation>
{%- endfor %}
{%- for bibtex_citation in bibtex_citations %}
            <citation type="bibtex">{{ bibtex_citation }}</citation>
{%- endfor %}
            <yield />
        </citations>
    </xml>
{%- if version_command %}
    <xml name="version_command">
        <version_command>{{ version_command }}</version_command>
    </xml>
{%- endif %}
</macros>
"""

CWL_TEMPLATE = """#!/usr/bin/env cwl-runner
cwlVersion: '{{cwl_version}}'
class: CommandLineTool
id: "{{id}}"
label: "{{label}}"
{%- if containers or requirements %}
hints:
{%- for container in containers %}
  DockerRequirement:
    dockerPull: {{ container.image_id }}
{%- endfor %}
{%- if requirements %}
  SoftwareRequirement:
    packages:
{%- for requirement in requirements %}
    - package: {{ requirement.name }}
{%- if requirement.version %}
      version:
      - "{{ requirement.version }}"
{%- else %}
      version: []
{%- endif %}
{%- endfor %}
{%- endif %}
{%- endif %}
{%- if inputs or outputs %}
inputs:
{%- for input in inputs %}
  {{ input.id }}:
    type: {{ input.type }}
    doc: |
      TODO
    inputBinding:
      position: {{ input.position }}
{%- if input.prefix %}
      prefix: "{{input.prefix.prefix}}"
{%- if not input.prefix.separated %}
      separate: false
{%- endif %}
{%- endif %}
{%- endfor %}
{%- for output in outputs %}
{%- if output.require_filename %}
  {{ output.id }}:
    type: string
    doc: |
      Filename for output {{ output.id }}
    inputBinding:
      position: {{ output.position }}
{%- if output.prefix %}
      prefix: "{{output.prefix.prefix}}"
{%- if not output.prefix.separated %}
      separate: false
{%- endif %}
{%- endif %}
{%- endif %}
{%- endfor %}
{%- else %}
inputs: [] # TODO
{%- endif %}
{%- if outputs %}
outputs:
{%- for output in outputs %}
  {{ output.id }}:
    type: File
    outputBinding:
      glob: {{ output.glob }}
{%- endfor %}
{%- else %}
outputs: [] # TODO
{%- endif %}
{%- if base_command %}
baseCommand:
{%- for base_command_part in base_command %}
  - "{{ base_command_part}}"
{%- endfor %}
{%- else %}
baseCommand: []
{%- endif %}
{%- if arguments %}
arguments:
{%- for argument in arguments %}
  - valueFrom: "{{ argument.value }}"
    position: {{ argument.position }}
{%- if argument.prefix %}
      prefix: "{{argument.prefix.prefix}}"
{%- if not argument.prefix.separated %}
      separate: false
{%- endif %}
{%- endif %}
{%- endfor %}
{%- else %}
arguments: []
{%- endif %}
{%- if stdout %}
stdout: {{ stdout }}
{%- endif %}
doc: |
{%- if help %}
  {{ help|indent(2) }}
{%- else %}
   TODO: Fill in description.
{%- endif %}
"""

CWL_TEST_TEMPLATE = """
- doc: test generated from example command
  job: {{ job_filename }}
{%- if outputs %}
  outputs:
{%- for output in outputs %}
    {{ output.id }}:
      path: test-data/{{ output.example_value }}
{%- endfor %}
{%- else %}
  outputs: TODO
{%- endif %}
"""

CWL_JOB_TEMPLATE = """
{%- if inputs %}
{%- for input in inputs %}
{%- if input.type == "File" %}
{{ input.id }}:
  class: File
  path: test-data/{{ input.example_value }}
{%- else %}
  {{ input.id }}: {{ input.example_value }}
{%- endif %}
{%- endfor %}
{%- else %}
# TODO: Specify job input.
{}
{%- endif %}
"""


def build(**kwds):
    """Build up a :func:`ToolDescription` from supplid arguments."""
    if kwds.get("cwl"):
        builder = _build_cwl
    else:
        builder = _build_galaxy
    return builder(**kwds)


def _build_cwl(**kwds):
    _handle_help(kwds)
    _handle_requirements(kwds)
    assert len(kwds["containers"]) <= 1, kwds
    command_io = CommandIO(**kwds)
    render_kwds = {
        "cwl_version": DEFAULT_CWL_VERSION,
        "help": kwds.get("help", ""),
        "containers": kwds.get("containers", []),
        "requirements": kwds.get("requirements", []),
        "id": kwds.get("id"),
        "label": kwds.get("name"),
    }
    render_kwds.update(command_io.cwl_properties())

    contents = _render(render_kwds, template_str=CWL_TEMPLATE)
    tool_files = []
    test_files = []
    if kwds["test_case"]:
        sep = "-" if "-" in kwds.get("id") else "_"
        tests_path = f"{kwds.get('id')}{sep}tests.yml"
        job_path = f"{kwds.get('id')}{sep}job.yml"
        render_kwds["job_filename"] = job_path
        test_contents = _render(render_kwds, template_str=CWL_TEST_TEMPLATE)
        job_contents = _render(render_kwds, template_str=CWL_JOB_TEMPLATE)
        tool_files.append(ToolFile(tests_path, test_contents, "test"))
        tool_files.append(ToolFile(job_path, job_contents, "job"))
        for cwl_input in render_kwds["inputs"] or []:
            if cwl_input.type == "File" and cwl_input.example_value:
                test_files.append(cwl_input.example_value)

        for cwl_output in render_kwds["outputs"] or []:
            if cwl_output.example_value:
                test_files.append(cwl_output.example_value)

    return ToolDescription(contents, tool_files=tool_files, test_files=test_files)


def _build_galaxy(**kwds):
    # Test case to build up from supplied inputs and outputs, ultimately
    # ignored unless kwds["test_case"] is truthy.

    _handle_help(kwds)

    # process raw cite urls
    cite_urls = kwds.get("cite_url", [])
    del kwds["cite_url"]
    citations = list(map(UrlCitation, cite_urls))
    kwds["bibtex_citations"] = citations

    # handle requirements and containers
    _handle_requirements(kwds)

    command_io = CommandIO(**kwds)
    kwds["inputs"] = command_io.inputs
    kwds["outputs"] = command_io.outputs
    kwds["command"] = command_io.cheetah_template

    test_case = command_io.test_case()

    # finally wrap up tests
    tests, test_files = _handle_tests(kwds, test_case)
    kwds["tests"] = tests

    # Render tool content from template.
    contents = _render(kwds)

    tool_files = []
    append_macro_file(tool_files, kwds)

    return ToolDescription(contents, tool_files=tool_files, test_files=test_files)


def append_macro_file(tool_files, kwds):
    macro_contents = None
    if kwds["macros"]:
        macro_contents = _render(kwds, MACROS_TEMPLATE)

        macros_file = "macros.xml"
        if not os.path.exists(macros_file):
            tool_files.append(ToolFile(macros_file, macro_contents, "macros"))

        io.info(REUSING_MACROS_MESSAGE)


class CommandIO:
    def __init__(self, **kwds):
        command = _find_command(kwds)
        cheetah_template = command

        # process raw inputs
        inputs = kwds.pop("input", [])
        inputs = list(map(Input, inputs or []))

        # alternatively process example inputs
        example_inputs = kwds.pop("example_input", [])
        for i, input_file in enumerate(example_inputs or []):
            name = f"input{i + 1}"
            inputs.append(Input(input_file, name=name, example=True))
            cheetah_template = _replace_file_in_command(cheetah_template, input_file, name)

        # handle raw outputs (from_work_dir ones) as well as named_outputs
        outputs = kwds.pop("output", [])
        outputs = list(map(Output, outputs or []))

        named_outputs = kwds.pop("named_output", [])
        for named_output in named_outputs or []:
            outputs.append(Output(name=named_output, example=False))

        # handle example outputs
        example_outputs = kwds.pop("example_output", [])
        for i, output_file in enumerate(example_outputs or []):
            name = f"output{i + 1}"
            from_path = output_file
            use_from_path = True
            if output_file in cheetah_template:
                # Actually found the file in the command, assume it can
                # be specified directly and skip from_work_dir.
                use_from_path = False
            output = Output(name=name, from_path=from_path, use_from_path=use_from_path, example=True)
            outputs.append(output)
            cheetah_template = _replace_file_in_command(cheetah_template, output_file, output.name)

        self.inputs = inputs
        self.outputs = outputs
        self.command = command
        self.cheetah_template = cheetah_template

    def example_input_names(self):
        for input in self.inputs:
            if input.example:
                yield input.input_description

    def example_output_names(self):
        for output in self.outputs:
            if output.example:
                yield output.example_path

    def cwl_lex_list(self):
        if not self.command:
            return []

        command_parts = shlex.split(self.command)
        parse_list = []

        input_count = 0
        output_count = 0

        index = 0

        prefixed_parts = []
        while index < len(command_parts):
            value = command_parts[index]
            eq_split = value.split("=")

            prefix = None
            if not _looks_like_start_of_prefix(index, command_parts):
                index += 1
            elif len(eq_split) == 2:
                prefix = Prefix(eq_split[0] + "=", False)
                value = eq_split[1]
                index += 1
            else:
                prefix = Prefix(value, True)
                value = command_parts[index + 1]
                index += 2
            prefixed_parts.append((prefix, value))

        for position, (prefix, value) in enumerate(prefixed_parts):
            if value in self.example_input_names():
                input_count += 1
                input = _CwlInput(
                    f"input{input_count}",
                    position,
                    prefix,
                    value,
                )
                parse_list.append(input)
            elif value in self.example_output_names():
                output_count += 1
                output = _CwlOutput(
                    f"output{output_count}",
                    position,
                    prefix,
                    value,
                )
                parse_list.append(output)
            elif prefix:
                param_id = prefix.prefix.lower().rstrip("=")
                type_ = param_type(value)
                input = _CwlInput(
                    param_id,
                    position,
                    prefix,
                    value,
                    type_=type_,
                )
                parse_list.append(input)
            else:
                part = _CwlCommandPart(value, position, prefix)
                parse_list.append(part)
        return parse_list

    def cwl_properties(self):
        base_command = []
        arguments = []
        inputs = []
        outputs = []

        lex_list = self.cwl_lex_list()

        index = 0
        while index < len(lex_list):
            token = lex_list[index]
            if isinstance(token, _CwlCommandPart):
                base_command.append(token.value)
            else:
                break
            index += 1

        while index < len(lex_list):
            token = lex_list[index]
            if token.is_token(">"):
                break
            token.position = index - len(base_command) + 1
            if isinstance(token, _CwlCommandPart):
                arguments.append(token)
            elif isinstance(token, _CwlInput):
                inputs.append(token)
            elif isinstance(token, _CwlOutput):
                token.glob = f"$(inputs.{token.id})"
                outputs.append(token)

            index += 1

        stdout = None
        if index < len(lex_list):
            token = lex_list[index]
            if token.is_token(">") and (index + 1) < len(lex_list):
                output_token = lex_list[index + 1]
                if not isinstance(output_token, _CwlOutput):
                    output_token = _CwlOutput("std_out", None)

                output_token.glob = "out"
                output_token.require_filename = False
                outputs.append(output_token)
                stdout = "out"
                index += 2
            else:
                io.warn("Example command too complex, you will need to build it up manually.")

        return {
            "inputs": inputs,
            "outputs": outputs,
            "arguments": arguments,
            "base_command": base_command,
            "stdout": stdout,
        }

    def test_case(self):
        test_case = TestCase()
        for input in self.inputs:
            if input.example:
                test_case.params.append((input.name, input.input_description))

        for output in self.outputs:
            if output.example:
                test_case.outputs.append((output.name, output.example_path))

        return test_case


def _looks_like_start_of_prefix(index, parts):
    value = parts[index]
    if len(value.split("=")) == 2:
        return True
    if index + 1 == len(parts):
        return False
    next_value = parts[index + 1]
    next_value_is_not_start = (len(value.split("=")) != 2) and next_value[0] not in ["-", ">", "<", "|"]
    return value.startswith("-") and next_value_is_not_start


Prefix = namedtuple("Prefix", ["prefix", "separated"])


class _CwlCommandPart:
    def __init__(self, value, position, prefix):
        self.value = value
        self.position = position
        self.prefix = prefix

    def is_token(self, value):
        return self.value == value


class _CwlInput:
    def __init__(self, id, position, prefix, example_value, type_="File"):
        self.id = id
        self.position = position
        self.prefix = prefix
        self.example_value = example_value
        self.type = type_

    def is_token(self, value):
        return False


class _CwlOutput:
    def __init__(self, id, position, prefix, example_value):
        self.id = id
        self.position = position
        self.prefix = prefix
        self.glob = None
        self.example_value = example_value
        self.require_filename = True

    def is_token(self, value):
        return False


def _render(kwds, template_str=TOOL_TEMPLATE):
    """Render template variables to generate the final tool."""
    return templates.render(template_str, **kwds)


def _replace_file_in_command(command, specified_file, name):
    """Replace example file with cheetah variable name.

    Be sure to single quote the name.
    """
    # TODO: check if the supplied variant was single quoted already.
    if f'"{specified_file}"' in command:
        # Sample command already wrapped filename in double quotes
        command = command.replace(f'"{specified_file}"', f"'${name}'")
    elif (f" {specified_file} ") in (" " + command + " "):
        # In case of spaces, best to wrap filename in double quotes
        command = command.replace(specified_file, f"'${name}'")
    else:
        command = command.replace(specified_file, f"${name}")
    return command


def _handle_help(kwds):
    """Convert supplied help parameters into a help variable for template.

    If help_text is supplied, use as is. If help is specified from a command,
    run the command and use that help text.
    """
    help_text = kwds.get("help_text")
    if not help_text:
        help_from_command = kwds.get("help_from_command")
        if help_from_command:
            p = subprocess.Popen(
                help_from_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True
            )
            help_text = p.communicate()[0]

    del kwds["help_text"]
    del kwds["help_from_command"]

    kwds["help"] = help_text


def _handle_tests(kwds, test_case):
    """Build tool test abstractions.

    Given state built up from handling rest of arguments (test_case) and
    supplied kwds - build tests for template and corresponding test files.
    """
    test_files = []
    if kwds["test_case"]:
        tests = [test_case]
        test_files.extend(map(lambda x: x[1], test_case.params))
        test_files.extend(map(lambda x: x[1], test_case.outputs))
    else:
        tests = []
    return tests, test_files


def _handle_requirements(kwds):
    """Build tool requirement abstractions.

    Convert requirements and containers specified from the command-line
    into abstract format for consumption by the template.
    """
    requirements = kwds["requirement"]
    del kwds["requirement"]
    requirements = list(map(Requirement, requirements or []))

    container = kwds["container"]
    del kwds["container"]
    containers = list(map(Container, container or []))

    kwds["requirements"] = requirements
    kwds["containers"] = containers


def _find_command(kwds):
    """Find base command from supplied arguments or just return `None`.

    If no such command was supplied (template will just replace this
    with a TODO item).
    """
    command = kwds.get("command")
    if not command:
        command = kwds.get("example_command", None)
        if command:
            del kwds["example_command"]
    return command


class UrlCitation:
    """Describe citation for tool."""

    def __init__(self, url):
        self.url = url

    def __str__(self):
        if "github.com" in self.url:
            return self._github_str()
        else:
            return self._url_str()

    def _github_str(self):
        url = self.url
        title = url.split("/")[-1]
        return f"""
@misc{{github{title},
  author = {{LastTODO, FirstTODO}},
  year = {{TODO}},
  title = {{{title}}},
  publisher = {{GitHub}},
  journal = {{GitHub repository}},
  url = {{{url}}},
}}"""

    def _url_str(self):
        url = self.url
        return f"""
@misc{{renameTODO,
  author = {{LastTODO, FirstTODO}},
  year = {{TODO}},
  title = {{TODO}},
  url = {{{url}}},
}}"""


class ToolDescription:
    """An description of the tool and related files to create."""

    def __init__(self, contents, tool_files=None, test_files=None):
        self.contents = contents
        self.tool_files = tool_files or []
        self.test_files = test_files or []


class ToolFile:
    def __init__(self, filename, contents, description):
        self.filename = filename
        self.contents = contents
        self.description = description


class Input:
    def __init__(self, input_description, name=None, example=False):
        parts = input_description.split(".")
        name = name or parts[0]
        if len(parts) > 0:
            datatype = ".".join(parts[1:])
        else:
            datatype = "data"

        self.input_description = input_description
        self.example = example
        self.name = name
        self.datatype = datatype

    def __str__(self):
        self.datatype = self.datatype.split(".")[-1]
        return f'<param type="data" name="{self.name}" format="{self.datatype}" />'


class Output:
    def __init__(self, from_path=None, name=None, use_from_path=False, example=False):
        if from_path:
            parts = from_path.split(".")
            name = name or parts[0]
            if len(parts) > 1:
                datatype = ".".join(parts[1:])
            else:
                datatype = "data"
        else:
            name = name
            datatype = "data"

        self.name = name
        self.datatype = datatype
        if use_from_path:
            self.from_path = from_path
        else:
            self.from_path = None
        self.example = example
        if example:
            self.example_path = from_path

    def __str__(self):
        if self.from_path:
            return self._from_path_str()
        else:
            return self._named_str()

    def _from_path_str(self):
        return f'<data name="{self.name}" format="{self.datatype}" from_work_dir="{self.from_path}" />'

    def _named_str(self):
        return f'<data name="{self.name}" format="{self.datatype}" />'


class Requirement:
    def __init__(self, requirement):
        parts = requirement.split("@", 1)
        if len(parts) > 1:
            name = parts[0]
            version = "@".join(parts[1:])
        else:
            name = parts[0]
            version = None
        self.name = name
        self.version = version

    def __str__(self):
        if self.version is not None:
            attrs = f' version="{self.version}"'
        else:
            attrs = ""
        return f'<requirement type="package"{attrs}>{self.name}</requirement>'


def param_type(value):
    if re.match(r"^\d+$", value):
        return "int"
    elif re.match(r"^\d+?\.\d+?$", value):
        return "float"
    else:
        return "string"


class Container:
    def __init__(self, image_id):
        self.type = "docker"
        self.image_id = image_id

    def __str__(self):
        return f'<container type="{self.type}">{self.image_id}</container>'


class TestCase:
    def __init__(self):
        self.params = []
        self.outputs = []


def write_tool_description(ctx, tool_description, **kwds):
    """Write a tool description to the file system guided by supplied CLI kwds."""
    tool_id = kwds.get("id")
    output = kwds.get("tool")
    if not output:
        extension = "cwl" if kwds.get("cwl") else "xml"
        output = f"{tool_id}.{extension}"
    if not io.can_write_to_path(output, **kwds):
        ctx.exit(1)

    io.write_file(output, tool_description.contents)
    io.info(f"Tool written to {output}")
    for tool_file in tool_description.tool_files:
        if tool_file.contents is None:
            continue

        path = tool_file.filename
        if not io.can_write_to_path(path, **kwds):
            ctx.exit(1)
        io.write_file(path, tool_file.contents)
        io.info(f"Tool {tool_file.description} written to {path}")

    macros = kwds["macros"]
    macros_file = "macros.xml"
    if macros and not os.path.exists(macros_file):
        io.write_file(macros_file, tool_description.macro_contents)
    elif macros:
        io.info(REUSING_MACROS_MESSAGE)
    if tool_description.test_files:
        if not os.path.exists("test-data"):
            io.info("No test-data directory, creating one.")
            os.makedirs("test-data")
        for test_file in tool_description.test_files:
            io.info(f"Copying test-file {test_file}")
            try:
                shutil.copy(test_file, "test-data")
            except Exception as e:
                io.info(f"Copy of {test_file} failed: {e}")


__all__ = (
    "build",
    "ToolDescription",
    "write_tool_description",
)



"""Planemo I/O abstractions and utilities."""

import contextlib
import errno
import fnmatch
import os
import shutil
import subprocess
import sys
import tempfile
import time
from io import StringIO
from sys import platform as _platform
from xml.sax.saxutils import escape

import click
from galaxy.util import commands
from galaxy.util.commands import download_command

from .exit_codes import (
    EXIT_CODE_NO_SUCH_TARGET,
    EXIT_CODE_OK,
)

IS_OS_X = _platform == "darwin"


def args_to_str(args):
    """Collapse list of arguments in a commmand-line string."""
    if args is None or isinstance(args, str):
        return args
    else:
        return commands.argv_to_str(args)


def communicate(cmds, **kwds):
    """Execute shell command and wait for output.

    With click-aware I/O handling, pretty display of the command being executed,
    and formatted exception if the exit code is not 0.
    """
    cmd_string = args_to_str(cmds)
    info(cmd_string)
    p = commands.shell_process(cmds, **kwds)
    if kwds.get("stdout", None) is None and commands.redirecting_io(sys=sys):
        output = commands.redirect_aware_commmunicate(p)
    else:
        output = p.communicate()

    if p.returncode != 0:
        msg = f"Problem executing commands {cmd_string} - ({output[0]}, {output[1]})"
        raise RuntimeError(msg)
    return output


def shell(cmds, **kwds):
    """Print and execute shell command."""
    cmd_string = args_to_str(cmds)
    info(cmd_string)
    return commands.shell(cmds, **kwds)


def info(message, *args):
    """Print stylized info message to the screen."""
    if args:
        message = message % args
    click.echo(click.style(message, bold=True, fg="green"))


def error(message, *args):
    """Print stylized error message to the screen."""
    if args:
        message = message % args
    click.echo(click.style(message, bold=True, fg="red"), err=True)


def warn(message, *args):
    """Print stylized warning message to the screen."""
    if args:
        message = message % args
    click.echo(click.style(message, fg="red"), err=True)


def can_write_to_path(path: str, **kwds):
    """Implement -f/--force logic.

    If supplied path exists, print an error message and return False
    unless --force caused the 'force' keyword argument to be True.
    """
    if not kwds["force"] and os.path.exists(path):
        error(f"{path} already exists, exiting.")
        return False
    return True


def shell_join(*args):
    """Join potentially empty commands together with '&&'."""
    return " && ".join(args_to_str(_) for _ in args if _)


def write_file(path, content, force=True):
    if os.path.exists(path) and not force:
        return

    with open(path, "w") as f:
        f.write(content)


def untar_to(url, tar_args=None, path=None, dest_dir=None):
    if tar_args:
        assert not (path and dest_dir)
        if dest_dir:
            if not os.path.exists(dest_dir):
                os.makedirs(dest_dir)
            tar_args[0:0] = ["-C", dest_dir]
        if path:
            tar_args.insert(0, "-O")

        download_cmd = download_command(url)
        download_p = commands.shell_process(download_cmd, stdout=subprocess.PIPE)
        untar_cmd = ["tar"] + tar_args
        if path:
            with open(path, "wb") as fh:
                shell(untar_cmd, stdin=download_p.stdout, stdout=fh)
        else:
            shell(untar_cmd, stdin=download_p.stdout)
        download_p.wait()
    else:
        cmd = download_command(url, to=path)
        shell(cmd)


def find_matching_directories(path, pattern, recursive):
    """Find directories below supplied path with file matching pattern.

    Returns an empty list if no matches are found, and if recursive is False
    only the top directory specified by path will be considered.
    """
    dirs = []
    if recursive:
        if not os.path.isdir(path):
            message = f"--recursive specified with non-directory path [{path}]"
            raise Exception(message)

        for base_path, dirnames, filenames in os.walk(path):
            dirnames.sort()
            for filename in fnmatch.filter(filenames, pattern):
                dirs.append(base_path)
    else:
        if os.path.exists(os.path.join(path, pattern)):
            dirs.append(path)
        elif os.path.basename(path) == pattern:
            dirs.append(os.path.dirname(path))
    return dirs


@contextlib.contextmanager
def real_io():
    """Ensure stdout and stderr have supported ``fileno()`` method.

    nosetests replaces these streams with :class:`StringIO` objects
    that may not work the same in every situtation - :func:`subprocess.Popen`
    calls in particular.
    """
    original_stdout = sys.stdout
    original_stderr = sys.stderr
    try:
        if commands.redirecting_io(sys=sys):
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__
        yield
    finally:
        sys.stdout = original_stdout
        sys.stderr = original_stderr


@contextlib.contextmanager
def temp_directory(prefix="planemo_tmp_", dir=None, **kwds):
    if dir is not None:
        try:
            os.makedirs(dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    temp_dir = tempfile.mkdtemp(prefix=prefix, dir=dir, **kwds)
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir)


def ps1_for_path(path, base="PS1"):
    """Used by environment commands to build a PS1 shell
    variables for tool or directory of tools.
    """
    file_name = os.path.basename(path)
    base_name = os.path.splitext(file_name)[0]
    ps1 = f"({base_name})${{{base}}}"
    return ps1


def stop_gravity(virtual_env, gravity_state_dir, env):
    gravity_bin = os.path.join(virtual_env, "bin", "galaxyctl")
    environ = os.environ.copy()
    environ.update(env)
    subprocess.check_call([gravity_bin, "--state-dir", gravity_state_dir, "shutdown"], env=environ)


def kill_pid_file(pid_file: str):
    """Kill process group corresponding to specified pid file."""
    try:
        os.stat(pid_file)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return False

    with open(pid_file) as fh:
        pid = int(fh.read())
    kill_posix(pid)
    try:
        os.unlink(pid_file)
    except Exception:
        pass


def kill_posix(pid: int):
    """Kill process group corresponding to specified pid."""

    def _check_pid():
        try:
            os.kill(pid, 0)
            return True
        except OSError:
            return False

    if _check_pid():
        for sig in [15, 9]:
            try:
                # gunicorn (unlike paste), seem to require killing process
                # group
                os.killpg(os.getpgid(pid), sig)
            except OSError:
                return
            time.sleep(1)
            if not _check_pid():
                return


@contextlib.contextmanager
def conditionally_captured_io(capture, tee=False):
    """If capture is True, capture stdout and stderr for logging."""
    captured_std = []
    if capture:
        with _Capturing() as captured_std:
            yield captured_std
        if tee:
            tee_captured_output(captured_std)
    else:
        yield


@contextlib.contextmanager
def captured_io_for_xunit(kwds, captured_io):
    """Capture Planemo I/O and timing for outputting to an xUnit report."""
    captured_std = []
    with_xunit = kwds.get("report_xunit", False)
    with conditionally_captured_io(with_xunit, tee=True):
        time1 = time.time()
        yield
        time2 = time.time()

    if with_xunit:
        stdout = [escape(m["data"]) for m in captured_std if m["logger"] == "stdout"]
        stderr = [escape(m["data"]) for m in captured_std if m["logger"] == "stderr"]
        captured_io["stdout"] = stdout
        captured_io["stderr"] = stderr
        captured_io["time"] = time2 - time1
    else:
        captured_io["stdout"] = None
        captured_io["stderr"] = None
        captured_io["time"] = None


class _Capturing(list):
    """Function context which captures stdout/stderr

    This keeps planemo's codebase clean without requiring planemo to hold onto
    messages, or pass user-facing messages back at all. This could probably be
    solved by swapping planemo entirely to a logger and reading from/writing
    to that, but this is easier.

    This swaps sys.std{out,err} with StringIOs and then makes that output
    available.
    """

    # http://stackoverflow.com/a/16571630

    def __enter__(self):
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        sys.stdout = self._stringio_stdout = StringIO()
        sys.stderr = self._stringio_stderr = StringIO()
        return self

    def __exit__(self, *args):
        self.extend([{"logger": "stdout", "data": x} for x in self._stringio_stdout.getvalue().splitlines()])
        self.extend([{"logger": "stderr", "data": x} for x in self._stringio_stderr.getvalue().splitlines()])

        sys.stdout = self._stdout
        sys.stderr = self._stderr


def tee_captured_output(output):
    """tee captured standard output and standard error if needed.

    For messages captured with Capturing, send them to their correct
    locations so as to not interfere with normal user experience.
    """
    for message in output:
        # Append '\n' due to `splitlines()` above
        if message["logger"] == "stdout":
            sys.stdout.write(message["data"] + "\n")
        if message["logger"] == "stderr":
            sys.stderr.write(message["data"] + "\n")


def wait_on(function, desc, timeout=5, polling_backoff=0):
    """Wait on given function's readiness.

    Grow the polling interval incrementally by the polling_backoff.
    """
    delta = 0.25
    timing = 0
    while True:
        if timing > timeout:
            message = f"Timed out waiting on {desc}."
            raise Exception(message)
        timing += delta
        delta += polling_backoff
        value = function()
        if value is not None:
            return value
        time.sleep(delta)


@contextlib.contextmanager
def open_file_or_standard_output(path, *args, **kwds):
    """Open file but respect '-' as referring to stdout."""
    if path == "-":
        yield sys.stdout
    else:
        yield open(path, *args, **kwds)


def filter_paths(paths, cwd=None, **kwds):
    if cwd is None:
        cwd = os.getcwd()

    def norm(path):
        if not os.path.isabs(path):
            path = os.path.join(cwd, path)
        return os.path.normpath(path)

    def exclude_func(exclude_path):
        def path_startswith(p):
            """Check that p starts with exclude_path and that the first
            character of p not included in exclude_path (if any) is the
            directory separator.
            """
            norm_p = norm(p)
            norm_exclude_path = norm(exclude_path)
            if norm_p.startswith(norm_exclude_path):
                return norm_p[len(norm_exclude_path) : len(norm_exclude_path) + 1] in ["", os.sep]
            return False

        return path_startswith

    filters_as_funcs = []
    filters_as_funcs.extend(map(exclude_func, kwds.get("exclude", [])))

    for exclude_paths_ins in kwds.get("exclude_from", []):
        with open(exclude_paths_ins) as f:
            for line in f.readlines():
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                filters_as_funcs.append(exclude_func(line))

    return [p for p in paths if not any(f(p) for f in filters_as_funcs)]


def coalesce_return_codes(ret_codes, assert_at_least_one=False):
    # Return 0 if everything is fine, otherwise pick the least
    # specific non-0 return code - preferring to report errors
    # to other non-0 exit codes.
    if assert_at_least_one and len(ret_codes) == 0:
        return EXIT_CODE_NO_SUCH_TARGET

    coalesced_return_code = EXIT_CODE_OK
    for ret_code in ret_codes:
        # None is equivalent to 0 in these methods.
        ret_code = 0 if ret_code is None else ret_code
        if ret_code == 0:
            # Everything is fine, keep moving...
            pass
        elif coalesced_return_code == 0:
            coalesced_return_code = ret_code
        # At this point in logic both ret_code and coalesced_return_code are
        # are non-zero
        elif ret_code < 0:
            # Error state, this should override eveything else.
            coalesced_return_code = ret_code
        elif ret_code > 0 and coalesced_return_code < 0:
            # Keep error state recorded.
            pass
        elif ret_code > 0:
            # Lets somewhat arbitrarily call the smaller exit code
            # the less specific.
            coalesced_return_code = min(ret_code, coalesced_return_code)

    if coalesced_return_code < 0:
        # Map -1 => 254, -2 => 253, etc...
        # Not sure it is helpful to have negative error codes
        # this was a design and API mistake in planemo.
        coalesced_return_code = 255 + coalesced_return_code

    return coalesced_return_code



"""Numeric constants for planemo exit codes."""

# Operation succeeded.
EXIT_CODE_OK = 0

# Generic failure.
EXIT_CODE_GENERIC_FAILURE = 1

# If iterating or tools for instance, not tool xml files are found.
EXIT_CODE_NO_SUCH_TARGET = 2

# An unknown file type was attempted to be processed with planemo.
EXIT_CODE_UNKNOWN_FILE_TYPE = 4

# An unsupported file type was supplied for a given operation.
EXIT_CODE_UNSUPPORTED_FILE_TYPE = 5

# A dependency of this operation was unavailable (e.g. conda).
EXIT_CODE_FAILED_DEPENDENCIES = 6

# Attempt to do a one time action that already happened.
EXIT_CODE_ALREADY_EXISTS = 7


class ExitCodeException(Exception):
    """Exception used by planemo framework to track exit codes for CLI."""

    def __init__(self, exit_code):
        """Specify integer exit code to exit planemo with."""
        self.exit_code = exit_code



"""Module describing the planemo ``create_gist`` command."""
import click

from planemo import github_util
from planemo.cli import command_function
from planemo.io import info

target_path = click.Path(
    file_okay=True,
    dir_okay=False,
    resolve_path=True,
)


@click.command("create_gist")
@click.argument(
    "path",
    metavar="FILE_PATH",
    type=target_path,
)
@click.option(
    "--link_type", type=click.Choice(["raw", "html"]), default="raw", help=("Link type to generate for the file.")
)
@command_function
def cli(ctx, path, **kwds):
    """Upload file to GitHub as a sharable gist."""
    file_url = github_util.publish_as_gist_file(ctx, path)
    if kwds.get("link_type") == "raw":
        share_url = file_url
    else:
        share_url = "http://htmlpreview.github.io/?%s" % file_url
    info("File published to Github Gist - share with %s" % share_url)



__version__ = "0.75.0"


PROJECT_NAME = "planemo"
PROJECT_OWNER = PROJECT_USERAME = "galaxyproject"
PROJECT_AUTHOR = "Galaxy Project and Community"
PROJECT_EMAIL = "jmchilton@gmail.com"

PROJECT_URL = "https://github.com/galaxyproject/planemo"
RAW_CONTENT_URL = f"https://raw.github.com/{PROJECT_USERAME}/{PROJECT_NAME}/master/"



"""Module describing the planemo ``ci_find_repos`` command."""

import click

from planemo import options
from planemo.ci import (
    filter_paths,
    print_path_list,
)
from planemo.cli import command_function
from planemo.shed import find_raw_repositories


@click.command("ci_find_repos")
@options.shed_project_arg()
@options.ci_find_options()
@command_function
def cli(ctx, paths, **kwds):
    """Find all shed repositories in one or more directories.

    Currently, a repository is considered any directory with a .shed.yml
    or .dockstore.yml file.
    """
    kwds["recursive"] = True
    kwds["fail_fast"] = True
    repos = find_raw_repositories(ctx, paths, **kwds)
    # Since fail_fast is True, all repos are actual raw repo objects and
    # not exceptions.
    raw_paths = [r.path for r in repos]
    paths = filter_paths(ctx, raw_paths, path_type="repo", **kwds)
    print_path_list(paths, **kwds)



import os

import requests

from planemo.galaxy.profiles import translate_alias
from planemo.galaxy.workflows import GALAXY_WORKFLOWS_PREFIX
from planemo.tools import uri_to_path
from .runnable import (
    for_path,
    for_uri,
    GALAXY_TOOLS_PREFIX,
)


def for_runnable_identifier(ctx, runnable_identifier, kwds, temp_path=None, return_all=False):
    """Convert URI, path, or alias into Runnable."""
    # could be a URI, path, or alias
    current_profile = kwds.get("profile")
    runnable_identifier = translate_alias(ctx, runnable_identifier, current_profile)
    if not runnable_identifier.startswith(GALAXY_WORKFLOWS_PREFIX):
        runnable_identifier = uri_to_path(ctx, runnable_identifier)
    if os.path.exists(runnable_identifier):
        runnable = for_path(runnable_identifier, temp_path=temp_path, return_all=return_all)
    else:  # assume galaxy workflow or tool id
        if "/repos/" in runnable_identifier:
            runnable_identifier = f"{GALAXY_TOOLS_PREFIX}{runnable_identifier}"
        elif not runnable_identifier.startswith(GALAXY_WORKFLOWS_PREFIX):
            runnable_identifier = f"{GALAXY_WORKFLOWS_PREFIX}{runnable_identifier}"
        runnable = for_uri(runnable_identifier)
    return runnable


def for_runnable_identifiers(ctx, runnable_identifiers, kwds, temp_path=None):
    """Convert lists of URIs, paths, and/or aliases into Runnables."""
    runnables = []
    for r in runnable_identifiers:
        runnable = for_runnable_identifier(ctx, r, kwds, temp_path=temp_path, return_all=True)
        if isinstance(runnable, list):
            runnables.extend(runnable)
        else:
            runnables.append(runnable)
    return runnables


def install_args_list_to_runnables(ctx, install_args_list, kwds):
    runnables = []
    for repo in install_args_list:
        url = f'{repo["tool_shed_url"]}api/repositories/get_repository_revision_install_info'
        response = requests.get(
            url, params={"name": repo["name"], "owner": repo["owner"], "changeset_revision": repo["changeset_revision"]}
        )
        response.raise_for_status()
        install_info = response.json()
        repository_metadata = install_info[1]
        assert repository_metadata["model_class"] == "RepositoryMetadata"
        for tool in repository_metadata.get("valid_tools", []):
            runnable = for_runnable_identifier(ctx, tool["guid"], kwds)
            runnables.append(runnable)
    return runnables



"""Utilities related to reasoning about test data."""

import os


def find_test_data_directory(tool_paths, **kwds):
    path = "."
    if len(tool_paths) > 0:
        path = tool_paths[0]

    # Find test data directory associated with path.
    test_data = kwds.get("test_data", None)
    if test_data:
        return os.path.abspath(test_data)
    else:
        test_data = search_tool_path_for(path, "test-data")
        if test_data:
            return test_data


def search_tool_path_for(path, target, extra_paths=[]):
    """Check for presence of a target in different artifact directories."""
    if not os.path.isdir(path):
        tool_dir = os.path.dirname(path)
    else:
        tool_dir = path
    possible_dirs = [tool_dir, "."] + extra_paths
    for possible_dir in possible_dirs:
        possible_path = os.path.join(possible_dir, target)
        if os.path.exists(possible_path):
            return os.path.abspath(possible_path)
    return None


__all__ = ("search_tool_path_for",)



"""Package is responsible for managing planemo profile databases.

This package makes it very easy to create and destroy databases, therefore it
should not be used for production data - and should not even be connnected
to a production database server.
"""

from .factory import create_database_source

__all__ = ("create_database_source",)



"""Module describing the planemo ``database_delete`` command."""

import click

from planemo import options
from planemo.cli import command_function
from planemo.database import create_database_source


@click.command("database_delete")
@options.database_identifier_argument()
@options.profile_database_options()
@options.docker_config_options()
@command_function
def cli(ctx, identifier, **kwds):
    """Delete a *development* database.

    Currently the only implementation is postgres which will be managed with
    ``psql``.

    Planemo ``database_`` commands make it very easy to create and destroy
    databases, therefore it should not be used for production data - and it
    should not even be connnected to a production database server. Planemo
    is intended for development purposes only.

    Planemo will assume that it can manage and access postgres databases
    without specifying a password. This can be accomplished by configuring
    postgres to not required a password for the planemo user or by specifying
    a password in a ``.pgpass`` file.

    Planemo can be configured to not require a password for the planemo user in
    the postgres configuration file ``pg_hba.conf`` (on Ubuntu/Debian linux
    distros this file is in /etc/postgresql/<postgres_version>/main/ directory).
    Adding the following lines to that file will allow planemo and Galaxy to
    access the databases without a password.

    \b
        # "local" is for Unix domain socket connections only
        local   all   all                    trust
        # IPv4 local connections:
        host    all   all    127.0.0.1/32    trust
        # IPv6 local connections:
        host    all   all    ::1/128         trust

    More information on the ``pg_hda.conf`` configuration file can be found at
    http://www.postgresql.org/docs/9.3/static/auth-pg-hba-conf.html.

    Information on ``.pgpass`` files can be found at at the following location:
    http://www.postgresql.org/docs/9.4/static/libpq-pgpass.html. In Ubuntu and
    Debian distros - a postgres user likely already exists and its password can
    be set by setting up a file ``~/.pgpass`` file with the following contents.

    \b
        *:*:*:postgres:<postgres_password>
    """
    create_database_source(**kwds).delete_database(identifier)



"""Module describing the planemo ``workflow_job_init`` command."""
import os

import click
import yaml

from planemo import options
from planemo.cli import command_function
from planemo.galaxy.workflows import (
    get_workflow_from_invocation_id,
    job_template,
    new_workflow_associated_path,
)
from planemo.io import can_write_to_path


@click.command("workflow_job_init")
@options.required_workflow_arg()
@options.force_option()
@options.workflow_output_artifact()
@options.galaxy_url_option()
@options.galaxy_user_key_option()
@options.from_invocation()
@options.profile_option()
@command_function
def cli(ctx, workflow_identifier, output=None, **kwds):
    """Initialize a Galaxy workflow job description for supplied workflow.

    Be sure to your lint your workflow with ``workflow_lint`` before calling this
    to ensure inputs and outputs comply with best practices that make workflow
    testing easier.

    Jobs can be run with the planemo run command (``planemo run workflow.ga job.yml``).
    Planemo run works with Galaxy tools and CWL artifacts (both tools and workflows)
    as well so this command may be renamed to to job_init at something along those
    lines at some point.
    """
    if kwds["from_invocation"]:
        if not os.path.isdir("test-data"):
            ctx.log("Creating test-data directory.")
            os.makedirs("test-data")
        path_basename = get_workflow_from_invocation_id(
            workflow_identifier, kwds["galaxy_url"], kwds["galaxy_user_key"]
        )

    job = job_template(workflow_identifier, **kwds)

    if output is None:
        output = new_workflow_associated_path(
            path_basename if kwds["from_invocation"] else workflow_identifier, suffix="job"
        )
    if not can_write_to_path(output, **kwds):
        ctx.exit(1)
    with open(output, "w") as f_job:
        yaml.dump(job, f_job)



from .build_report import template_data


def handle_report_xunit_kwd(kwds, collected_data):
    if kwds.get("report_xunit", False):
        with open(kwds["report_xunit"], "w") as handle:
            handle.write(template_data(collected_data, template_name="xunit.tpl"))



"""Module describing the planemo ``project_init`` command."""
import os
import shutil
import tempfile

import click

from planemo import options
from planemo.cli import command_function
from planemo.io import (
    untar_to,
    warn,
)

SOURCE_HOST = "https://codeload.github.com"
DOWNLOAD_URL = "%s/galaxyproject/planemo/tar.gz/master" % SOURCE_HOST


@click.command("project_init")
@options.optional_project_arg(exists=None)
@click.option("--template", default=None)
@command_function
def cli(ctx, path, template=None, **kwds):
    """(Experimental) Initialize a new tool project.

    This is only a proof-of-concept demo right now.
    """
    if template is None:
        warn("Creating empty project, this function doesn't do much yet.")
    if not os.path.exists(path):
        os.makedirs(path)
    if template is None:
        return

    tempdir = tempfile.mkdtemp()
    tar_args = ["-zxf", "-", "--strip-components=2"]
    try:
        untar_to(DOWNLOAD_URL, tar_args=tar_args, dest_dir=tempdir)
        template_dir = os.path.join(tempdir, template)
        for entry in os.listdir(template_dir):
            shutil.move(os.path.join(template_dir, entry), path)
    finally:
        shutil.rmtree(tempdir)



"""Module contains factory method for building class:`Engine` objects."""

import contextlib
from typing import Generator

from planemo.engine.interface import BaseEngine
from .cwltool import CwlToolEngine
from .galaxy import (
    DockerizedManagedGalaxyEngine,
    ExternalGalaxyEngine,
    LocalManagedGalaxyEngine,
)
from .toil import ToilEngine

UNKNOWN_ENGINE_TYPE_MESSAGE = "Unknown engine type specified [%s]."


def is_galaxy_engine(**kwds):
    """Return True iff the engine configured is :class:`GalaxyEngine`."""
    engine_type_str = kwds.get("engine", "galaxy")
    return engine_type_str in ["galaxy", "docker_galaxy", "external_galaxy"]


def build_engine(ctx, **kwds):
    """Build an engine from the supplied planemo configuration."""
    engine_type_str = kwds.get("engine", "galaxy")
    if engine_type_str == "galaxy":
        engine_type = LocalManagedGalaxyEngine
    elif engine_type_str == "docker_galaxy":
        engine_type = DockerizedManagedGalaxyEngine
    elif engine_type_str == "external_galaxy":
        engine_type = ExternalGalaxyEngine
    elif engine_type_str == "cwltool":
        engine_type = CwlToolEngine
    elif engine_type_str == "toil":
        engine_type = ToilEngine
    else:
        raise Exception(UNKNOWN_ENGINE_TYPE_MESSAGE % engine_type_str)

    return engine_type(ctx, **kwds)


@contextlib.contextmanager
def engine_context(ctx, **kwds) -> Generator[BaseEngine, None, None]:
    """A :func:`contextlib.contextmanager` engine builder for use with ``with`` statements.

    https://docs.python.org/2/library/contextlib.html
    """
    engine = None
    try:
        engine = build_engine(ctx, **kwds)
        yield engine
    finally:
        if engine is not None:
            engine.cleanup()


__all__ = (
    "is_galaxy_engine",
    "build_engine",
    "engine_context",
)



"""Module describing the planemo ``profile_create`` command."""

import click

from planemo import options
from planemo.cli import command_function
from planemo.galaxy import profiles


@click.command("profile_create")
@options.profile_name_argument()
@options.profile_database_options()
@options.serve_engine_option()
@options.docker_config_options()
@options.galaxy_url_option()
@options.galaxy_user_key_option()
@options.galaxy_admin_key_option()
@command_function
def cli(ctx, profile_name, **kwds):
    """Create a profile."""
    profiles.create_profile(ctx, profile_name, **kwds)
    print("Profile [%s] created." % profile_name)



import inspect
import json
import os
import re
from collections import OrderedDict
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TYPE_CHECKING,
    Union,
)

import yaml
from galaxy.tool_util.lint import LintContext
from galaxy.tool_util.loader_directory import EXCLUDE_WALK_DIRS
from galaxy.tool_util.verify import asserts
from gxformat2.lint import (
    lint_format2,
    lint_ga,
)
from gxformat2.yaml import ordered_load

from planemo.exit_codes import (
    EXIT_CODE_GENERIC_FAILURE,
    EXIT_CODE_OK,
)
from planemo.galaxy.workflows import (
    input_labels,
    output_labels,
    required_input_labels,
)
from planemo.runnable import (
    cases,
    for_path,
    TestCase,
)
from planemo.shed import DOCKSTORE_REGISTRY_CONF

if TYPE_CHECKING:
    from planemo.cli import PlanemoCliContext

POTENTIAL_WORKFLOW_FILES = re.compile(r"^.*(\.yml|\.yaml|\.ga)$")
DOCKSTORE_REGISTRY_CONF_VERSION = "1.2"


class WorkflowLintContext(LintContext):
    # Setup training topic for linting - probably should pass this through
    # from click arguments.
    training_topic = None


def generate_dockstore_yaml(directory: str, publish: bool = True) -> str:
    workflows = []
    for workflow_path in find_workflow_descriptions(directory):
        test_parameter_path = f"{workflow_path.rsplit('.', 1)[0]}-tests.yml"
        workflow_entry: Dict[str, Any] = {
            # TODO: support CWL
            "subclass": "Galaxy",
            "publish": publish,
            "name": "main",
            "primaryDescriptorPath": f"/{os.path.relpath(workflow_path, directory)}",
        }
        if os.path.exists(test_parameter_path):
            workflow_entry["testParameterFiles"] = [f"/{os.path.relpath(test_parameter_path, directory)}"]
        workflows.append(workflow_entry)
    # Force version to the top of file but serializing rest of config separately
    contents = "version: %s\n" % DOCKSTORE_REGISTRY_CONF_VERSION
    contents += yaml.dump({"workflows": workflows})
    return contents


def lint_workflow_artifacts_on_paths(
    ctx: "PlanemoCliContext", paths: Iterable[str], lint_args: Dict[str, Union[str, List[str]]]
) -> int:
    report_level = lint_args["level"]
    lint_context = WorkflowLintContext(report_level, skip_types=lint_args["skip_types"])
    for path in paths:
        _lint_workflow_artifacts_on_path(lint_context, path, lint_args)

    if lint_context.failed(lint_args["fail_level"]):
        return EXIT_CODE_GENERIC_FAILURE
    else:
        return EXIT_CODE_OK


def _lint_workflow_artifacts_on_path(
    lint_context: WorkflowLintContext, path: str, lint_args: Dict[str, Union[str, List[str]]]
) -> None:
    for potential_workflow_artifact_path in find_potential_workflow_files(path):
        if os.path.basename(potential_workflow_artifact_path) == DOCKSTORE_REGISTRY_CONF:
            lint_context.lint("lint_dockstore", _lint_dockstore_config, potential_workflow_artifact_path)

        elif looks_like_a_workflow(potential_workflow_artifact_path):

            def structure(path, lint_context):
                with open(path) as f:
                    workflow_dict = ordered_load(f)
                workflow_class = workflow_dict.get("class")
                lint_func = lint_format2 if workflow_class == "GalaxyWorkflow" else lint_ga
                lint_func(lint_context, workflow_dict, path=path)

            lint_context.lint("lint_structure", structure, potential_workflow_artifact_path)
            lint_context.lint("lint_best_practices", _lint_best_practices, potential_workflow_artifact_path)
            lint_context.lint("lint_tests", _lint_tsts, potential_workflow_artifact_path)
        else:
            # Allow linting ro crates and such also
            pass


# misspell for pytest
def _lint_tsts(path: str, lint_context: WorkflowLintContext) -> None:
    runnables = for_path(path, return_all=True)
    if not isinstance(runnables, list):
        runnables = [runnables]
    for runnable in runnables:
        test_cases = cases(runnable)
        if len(test_cases) == 0:
            lint_context.warn("Workflow missing test cases.")
            return
        all_tests_valid = True
        for test_case in test_cases:
            if isinstance(test_case, TestCase):
                if not _lint_case(path, test_case, lint_context):
                    all_tests_valid = False
            else:
                lint_context.warn(f"Test case of type {type(test_case)} not currently supported.")
                all_tests_valid = False

        if all_tests_valid:
            lint_context.valid(f"Tests appear structurally correct for {runnable.path}")


def _lint_best_practices(path: str, lint_context: WorkflowLintContext) -> None:  # noqa: C901
    """
    This function duplicates the checks made by Galaxy's best practices panel:
    https://github.com/galaxyproject/galaxy/blob/5396bb15fe8cfcf2e89d46c1d061c49b60e2f0b1/client/src/components/Workflow/Editor/Lint.vue
    """

    def check_json_for_untyped_params(j):
        values = j if type(j) == list else j.values()
        for value in values:
            if type(value) in [list, dict, OrderedDict]:
                if check_json_for_untyped_params(value):
                    return True
            elif type(value) == str:
                if re.match(r"\$\{.+?\}", value):
                    return True
        return False

    with open(path) as f:
        workflow_dict = ordered_load(f)

    steps = workflow_dict.get("steps", {})

    # annotation
    if not workflow_dict.get("annotation"):
        lint_context.warn("Workflow is not annotated.")

    # creator
    if not len(workflow_dict.get("creator", [])) > 0:
        lint_context.warn("Workflow does not specify a creator.")

    # license
    if not workflow_dict.get("license"):
        lint_context.warn("Workflow does not specify a license.")

    # checks on individual steps
    for step in steps.values():
        print(step)
        # disconnected inputs
        for input in step.get("inputs", []):
            if input.get("name") not in step.get("input_connections"):  # TODO: check optional
                lint_context.warn(
                    f"Input {input.get('name')} of workflow step {step.get('annotation') or step.get('id')} is disconnected."
                )

        # missing metadata
        if not step.get("annotation"):
            lint_context.warn(f"Workflow step with ID {step.get('id')} has no annotation.")
        if not step.get("label"):
            lint_context.warn(f"Workflow step with ID {step.get('id')} has no label.")

        # untyped parameters
        if workflow_dict.get("class") == "GalaxyWorkflow":
            tool_state = step.get("tool_state", {})
            pjas = step.get("out", {})
        else:
            tool_state = json.loads(step.get("tool_state", "{}"))
            pjas = step.get("post_job_actions", {})

        if check_json_for_untyped_params(tool_state):
            lint_context.warn(f"Workflow step with ID {step.get('id')} specifies an untyped parameter as an input.")

        if check_json_for_untyped_params(pjas):
            lint_context.warn(
                f"Workflow step with ID {step.get('id')} specifies an untyped parameter in the post-job actions."
            )

        # unlabeled outputs are checked by gxformat2, no need to check here


def _lint_case(path: str, test_case: TestCase, lint_context: WorkflowLintContext) -> bool:
    test_valid = True

    i_labels = input_labels(workflow_path=path)
    job_keys = test_case.input_ids
    for key in job_keys:
        if key not in i_labels:
            # consider an error instead?
            lint_context.warn(
                f"Unknown workflow input in test job definition [{key}], workflow inputs are [{i_labels}]"
            )
            test_valid = False

    # check non-optional parameters are set
    for required_label in required_input_labels(path):
        if required_label not in job_keys:
            template = "Non-optional input has no value specified in workflow test job [%s], job specifies inputs [%s]"
            lint_context.error(template % (required_label, job_keys))
            test_valid = False

    for input_id, input_def in test_case._job.items():
        if not _tst_input_valid(test_case, input_id, input_def, lint_context):
            test_valid = False

    test_output_ids = test_case.tested_output_ids
    o_labels = output_labels(path)
    found_valid_expectation = False
    for test_output_id in test_output_ids:
        if test_output_id not in o_labels:
            template = "Test found for unknown workflow output [%s], workflow outputs [%s]"
            lint_context.error(template % (test_output_id, o_labels))
            test_valid = False
        else:
            found_valid_expectation = True
        # TODO: validate structure of test expectations

        assertion_definitions = test_case.output_expectations[test_output_id].get("asserts")
        if not _check_test_assertions(lint_context, assertion_definitions):
            test_valid = False

    if not found_valid_expectation:
        lint_context.warn("Found no valid test expectations for workflow test")
        test_valid = False

    return test_valid


def _check_test_assertions(
    lint_context: WorkflowLintContext, assertion_definitions: Optional[Dict[str, Dict[str, Any]]]
) -> bool:
    # we are already in Python, not XML, so it is simpler to lint assertions by checking against the
    # Python functions directly, rather than checking against galaxy.xsd as for tool linting
    assertions_valid = True
    if assertion_definitions:
        for assertion_name, assertion_params in assertion_definitions.items():
            function = asserts.assertion_functions[f"assert_{assertion_name}"]
            signature = inspect.signature(function)
            try:
                # try mapping the function with the attributes supplied and check for TypeError
                signature.bind("", **assertion_params)
            except AssertionError:
                pass
            except TypeError as e:
                lint_context.error(f"Invalid assertion in tests: assert_{assertion_name} {str(e)}")
                assertions_valid = False
    return assertions_valid


def _tst_input_valid(
    test_case: TestCase,
    input_id: str,
    input_def: Dict[str, Any],
    lint_context: WorkflowLintContext,
) -> bool:
    if type(input_def) == dict:  # else assume it is a parameter
        clazz = input_def.get("class")
        if clazz == "File":
            input_path = input_def.get("path")
            if input_path:
                if not os.path.isabs(input_path):
                    input_path = os.path.join(test_case.tests_directory, input_path)
                if not os.path.exists(input_path):
                    message = "Test referenced File path [%s] not found" % input_path
                    lint_context.warn(message)
                    return False
        elif clazz == "Collection":
            for elem in input_def.get("elements", []):
                elem_valid = _tst_input_valid(test_case, input_id, elem, lint_context)
                if not elem_valid:
                    return False
    return True


def _lint_dockstore_config(path: str, lint_context: WorkflowLintContext) -> None:
    dockstore_yaml = None
    try:
        with open(path) as f:
            dockstore_yaml = yaml.safe_load(f)
    except Exception:
        lint_context.error("Invalid YAML found in %s" % DOCKSTORE_REGISTRY_CONF)
        return

    if not isinstance(dockstore_yaml, dict):
        lint_context.error("Invalid YAML contents found in %s" % DOCKSTORE_REGISTRY_CONF)
        return

    if "workflows" not in dockstore_yaml:
        lint_context.error("Invalid YAML contents found in %s, no workflows defined" % DOCKSTORE_REGISTRY_CONF)
        return

    workflow_entries = dockstore_yaml.get("workflows")
    if not isinstance(workflow_entries, list):
        lint_context.error("Invalid YAML contents found in %s, workflows not a list" % DOCKSTORE_REGISTRY_CONF)
        return

    for workflow_entry in workflow_entries:
        _lint_dockstore_workflow_entry(lint_context, os.path.dirname(path), workflow_entry)


def _lint_dockstore_workflow_entry(
    lint_context: WorkflowLintContext, directory: str, workflow_entry: Dict[str, Any]
) -> None:
    if not isinstance(workflow_entry, dict):
        lint_context.error("Invalid YAML contents found in %s, workflow entry not a dict" % DOCKSTORE_REGISTRY_CONF)
        return

    found_errors = False
    for required_key in ["primaryDescriptorPath", "subclass"]:
        if required_key not in workflow_entry:
            lint_context.error(f"{DOCKSTORE_REGISTRY_CONF} workflow entry missing required key {required_key}")
            found_errors = True

    for recommended_key in ["testParameterFiles"]:
        if recommended_key not in workflow_entry:
            lint_context.warn(f"{DOCKSTORE_REGISTRY_CONF} workflow entry missing recommended key {recommended_key}")

    if found_errors:
        # Don't do the rest of the validation for a broken file.
        return

    # TODO: validate subclass
    descriptor_path = workflow_entry["primaryDescriptorPath"]
    test_files = workflow_entry.get("testParameterFiles", [])

    for referenced_file in [descriptor_path] + test_files:
        referenced_path = os.path.join(directory, referenced_file[1:])
        if not os.path.exists(referenced_path):
            lint_context.error(f"{DOCKSTORE_REGISTRY_CONF} workflow entry references absent file {referenced_file}")


def looks_like_a_workflow(path: str) -> bool:
    """Return boolean indicating if this path looks like a workflow."""
    if POTENTIAL_WORKFLOW_FILES.match(os.path.basename(path)):
        with open(path) as f:
            workflow_dict = ordered_load(f)
        if not isinstance(workflow_dict, dict):
            # Not exactly right - could have a #main def - do better and sync with Galaxy.
            return False
        return bool(workflow_dict.get("class") == "GalaxyWorkflow" or workflow_dict.get("a_galaxy_workflow"))
    return False


def find_workflow_descriptions(directory: str) -> Iterator[str]:
    for potential_workflow_artifact_path in find_potential_workflow_files(directory):
        if looks_like_a_workflow(potential_workflow_artifact_path):
            yield potential_workflow_artifact_path


def find_potential_workflow_files(directory: str) -> List[str]:
    """Return a list of potential workflow files in a directory."""
    if not os.path.exists(directory):
        raise ValueError(f"Directory not found {directory}")

    matches = []
    if os.path.isdir(directory):
        for root, dirnames, filenames in os.walk(directory):
            # exclude some directories (like .hg) from traversing
            dirnames[:] = [dir for dir in dirnames if dir not in EXCLUDE_WALK_DIRS]
            for filename in filenames:
                if POTENTIAL_WORKFLOW_FILES.match(filename):
                    matches.append(os.path.join(root, filename))
    else:
        matches.append(directory)
    return matches



"""Module describing the planemo ``pull_request`` command."""
import click

from planemo import (
    github_util,
    options,
)
from planemo.cli import command_function
from planemo.config import planemo_option


@click.command("pull_request")
@planemo_option(
    "-m", "--message", type=click.STRING, default=None, help="Message describing the pull request to create."
)
@options.optional_project_arg(exists=None)
@command_function
def cli(ctx, path, message=None, **kwds):
    """Short-cut to quickly create a pull request for a relevant Github repo.

    For instance, the following will clone, fork, and branch the tools-iuc
    repository to allow a subsequent pull request to fix a problem with bwa.

    \b
        $ planemo clone --branch bwa-fix tools-iuc
        $ cd tools-iuc
        $ # Make changes.
        $ git add -p # Add desired changes.
        $ git commit -m "Fix bwa problem."
        $ planemo pull_request -m "Fix bwa problem."

    These changes do require that a github access token is
    specified in ~/.planemo.yml. An access token can be generated by going
    to https://github.com/settings/tokens.
    """
    github_util.pull_request(ctx, path, message=message, **kwds)



"""Module describing the planemo ``run`` command."""

import json

import click
from galaxy.util import unicodify

from planemo import options
from planemo.cli import command_function
from planemo.engine import engine_context
from planemo.galaxy.test import handle_reports_and_summary
from planemo.io import (
    info,
    warn,
)
from planemo.runnable_resolve import for_runnable_identifier
from planemo.test.results import StructuredData


@click.command("run")
@options.required_runnable_arg()
@options.required_job_arg()
@options.galaxy_run_options()
@options.galaxy_config_options()
@options.enable_cwl_option()
@options.galaxy_cwl_root_option()
@options.run_history_tags_option()
@options.run_output_directory_option()
@options.run_output_json_option()
@options.run_download_outputs_option()
@options.engine_options()
@options.test_options()
@command_function
def cli(ctx, runnable_identifier, job_path, **kwds):
    """Planemo command for running tools and jobs.

    \b
        % planemo run cat1-tool.cwl cat-job.json
    """
    runnable = for_runnable_identifier(ctx, runnable_identifier, kwds)
    is_cwl = runnable.type.is_cwl_artifact
    kwds["cwl"] = is_cwl
    kwds["execution_type"] = "Run"
    if kwds.get("engine", None) is None:
        if is_cwl:
            kwds["engine"] = "cwltool"
        elif kwds.get("galaxy_url", None):
            kwds["engine"] = "external_galaxy"
        else:
            kwds["engine"] = "galaxy"
    with engine_context(ctx, **kwds) as engine:
        run_result = engine.run([runnable], [job_path])[0]

    if not run_result.was_successful:
        warn("Run failed [%s]" % unicodify(run_result))
    elif kwds.get("no_wait"):
        info("Run successfully executed - exiting without waiting for results.")
    else:
        output_json = kwds.get("output_json", None)
        outputs_dict = run_result.outputs_dict
        if output_json:
            with open(output_json, "w") as f:
                json.dump(outputs_dict, f)
        info("Run completed successfully.")

    report_data = StructuredData(data={"tests": [run_result.structured_data()], "version": "0.1"})
    report_data.calculate_summary_data()
    return_value = handle_reports_and_summary(ctx, report_data.structured_data, kwds=kwds)
    ctx.exit(return_value)



"""Module describing the planemo ``workflow_edit`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.engine import (
    engine_context,
    is_galaxy_engine,
)
from planemo.galaxy.serve import sleep_for_serve
from planemo.runnable_resolve import for_runnable_identifier


@click.command("workflow_edit")
@options.required_workflow_arg()
@options.galaxy_serve_options()
@command_function
def cli(ctx, workflow_identifier, output=None, force=False, **kwds):
    """Open a synchronized Galaxy workflow editor."""
    assert is_galaxy_engine(**kwds)
    runnable = for_runnable_identifier(ctx, workflow_identifier, kwds.get("profile"))

    kwds["workflows_from_path"] = True

    with engine_context(ctx, **kwds) as galaxy_engine:
        with galaxy_engine.ensure_runnables_served([runnable]) as config:
            workflow_id = config.workflow_id_for_runnable(runnable)
            url = f"{config.galaxy_url}/workflow/editor?id={workflow_id}"
            click.launch(url)
            if kwds["engine"] != "external_galaxy":
                sleep_for_serve()



"""Module describes a :class:`DatabaseSource` for local postgres databases."""

import subprocess

from galaxy.util import unicodify

from planemo.io import communicate
from .interface import DatabaseSource


class ExecutesPostgresSqlMixin:
    def list_databases(self):
        """Use `psql --list` to generate a list of identifiers."""
        command_builder = self._psql_command_builder("--list")
        stdout = unicodify(self._communicate(command_builder))
        output_lines = stdout.splitlines()
        identifiers = []
        for line in output_lines:
            identifiers.append(line.split("|")[0].strip())
        return [i for i in identifiers if i]

    def create_database(self, identifier):
        """Use `psql -c "create database"` to create a database."""
        sql = "create database %s;" % identifier
        self._run_sql_command(sql)

    def delete_database(self, identifier):
        """Use `psql -c "drop database"` to delete a database."""
        sql = "drop database %s;" % identifier
        self._run_sql_command(sql)

    def _run_sql_command(self, sql):
        # communicate is just joining commands so we need to modify the
        # sql as an argument - it shouldn't do this.
        sql_arg = "%s" % sql
        command_builder = self._psql_command_builder("--command", sql_arg)
        self._communicate(command_builder)

    def _communicate(self, command_builder):
        stdout, _ = communicate(
            command_builder.command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return stdout


class LocalPostgresDatabaseSource(ExecutesPostgresSqlMixin, DatabaseSource):
    """Local postgres database source managed through psql application."""

    def __init__(self, **kwds):
        """Construct a postgres database source from planemo configuration."""
        self.psql_path = kwds.get("postgres_psql_path", None) or "psql"
        self.database_user = kwds.get("postgres_database_user", None)
        self.database_host = kwds.get("postgres_database_host", None)
        self.database_port = kwds.get("postgres_database_port", None)
        self._kwds = kwds

    def sqlalchemy_url(self, identifier):
        """Return URL or form postgresql://username:password@localhost/mydatabase."""
        hostname = self.database_host or "localhost"
        if self.database_port:
            hostname += ":%s" % self.database_port
        return f"postgresql://{self.database_user}@{hostname}/{identifier}"

    def _psql_command_builder(self, *args):
        command_builder = _CommandBuilder(self.psql_path)
        # Print only tuples so output is easier to parse
        command_builder.append_command("--tuples-only")

        # Specify connection information
        if self.database_user:
            command_builder.append_command("--username", self.database_user)
        if self.database_host:
            command_builder.append_command("--host", self.database_host)
        if self.database_port:
            command_builder.append_command("--port", self.database_port)
        command_builder.append_command("-P", "pager=off")
        command_builder.extend_command(args)
        return command_builder


class _CommandBuilder:
    def __init__(self, *args):
        self.command = list(args)

    def append_command(self, *args_or_none):
        args_or_none = args_or_none or []
        for arg_or_none in args_or_none:
            if arg_or_none is not None:
                self.command.append(arg_or_none)

    def extend_command(self, args):
        for arg in args or []:
            self.append_command(arg)


__all__ = ("LocalPostgresDatabaseSource",)



"""Module describing the planemo ``share_test`` command."""
import click

from planemo import (
    github_util,
    options,
)
from planemo.cli import command_function
from planemo.io import info

PLANEMO_TEST_VIEWER_URL_TEMPLATE = "http://galaxyproject.github.io/planemo/tool_test_viewer.html" "?test_data_url=%s"


@click.command("share_test")
@options.tool_test_json()
@command_function
def cli(ctx, path, **kwds):
    """Publish JSON test results as sharable Gist.

    This will upload the JSON test results to Github as a Gist and
    produce sharable URL.

    The sharable URL can be used to share an HTML version of the report
    that can be easily embedded in pull requests or commit messages.

    Requires a ~/.planemo.yml with a Github access token
    defined in a 'github' section of that configuration file. An access
    token can be generated by going to https://github.com/settings/tokens.
    """
    file_url = github_util.publish_as_gist_file(ctx, path)
    share_url = PLANEMO_TEST_VIEWER_URL_TEMPLATE % file_url
    info("File published to Github Gist.")
    info("Raw URL: %s" % file_url)
    info("Share results with URL: %s" % share_url)
    markdown = "[View Tool Test Results](%s)" % share_url
    info("Embed results with markdown: %s" % markdown)



"""Module describing the planemo ``training_init`` command."""

import click

from planemo import options
from planemo.cli import command_function
from planemo.training import Training


@click.command("training_init")
@options.optional_tools_arg(multiple=True, allow_uris=True)
@options.training_init_options()
# @options.force_option()
@options.galaxy_serve_options()
@command_function
def cli(ctx, uris, **kwds):
    """Build training template from workflow."""
    kwds["no_dependency_resolution"] = True
    training = Training(kwds)
    training.init_training(ctx)



"""Planemo layer for ensuring bioblend available."""
try:
    from bioblend import (
        galaxy,
        toolshed,
    )
except ImportError:
    toolshed = None
    galaxy = None

BIOBLEND_UNAVAILABLE = (
    "This functionality requires the bioblend library "
    " which is unavailable, please install `pip install "
    "bioblend`"
)


def ensure_module() -> None:
    """Throw an exception if bioblend is not available to Planemo."""
    if toolshed is None:
        raise Exception(BIOBLEND_UNAVAILABLE)



"""Module describing the planemo ``conda_build`` command."""

import click

from planemo import options
from planemo.cli import command_function
from planemo.conda import build_conda_context


@click.command("conda_build")
@options.conda_target_options(include_local=False)  # No reason to expose local, we have to use it.
@options.recipe_arg(multiple=True)
@command_function
def cli(ctx, paths, **kwds):
    """Perform conda build with Planemo's conda."""
    # Force conda_use_local for building...
    kwds["conda_use_local"] = True
    conda_context = build_conda_context(ctx, handle_auto_init=True, **kwds)
    build_args = list(paths)
    conda_context.exec_command("build", build_args)



"""Logic related to linting shed repositories."""

import os
import xml.etree.ElementTree as ET

import yaml
from galaxy.tool_util.lint import lint_tool_source_with
from galaxy.tool_util.linters.help import rst_invalid
from galaxy.util import unicodify

from planemo.io import info
from planemo.lint import (
    handle_lint_complete,
    lint_urls,
    lint_xsd,
    setup_lint,
)
from planemo.shed import (
    CURRENT_CATEGORIES,
    REPO_TYPE_SUITE,
    REPO_TYPE_TOOL_DEP,
    REPO_TYPE_UNRESTRICTED,
    validate_repo_name,
    validate_repo_owner,
)
from planemo.shed2tap import base
from planemo.tool_lint import (
    build_tool_lint_args,
    handle_tool_load_error,
)
from planemo.tools import yield_tool_sources
from planemo.xml import XSDS_PATH

TOOL_DEPENDENCIES_XSD = os.path.join(XSDS_PATH, "tool_dependencies.xsd")
REPO_DEPENDENCIES_XSD = os.path.join(XSDS_PATH, "repository_dependencies.xsd")


VALID_REPOSITORY_TYPES = [
    REPO_TYPE_UNRESTRICTED,
    REPO_TYPE_TOOL_DEP,
    REPO_TYPE_SUITE,
]

SHED_METADATA = [
    "description",
    "long_description",
    "remote_repository_url",
    "homepage_url",
    "categories",
]


def lint_repository(ctx, realized_repository, **kwds):
    """Lint a realized shed repository.

    See :mod:`planemo.shed` for details on constructing a realized
    repository data structure.
    """
    failed = False
    path = realized_repository.real_path
    info("Linting repository %s" % path)
    lint_args = build_tool_lint_args(ctx, **kwds)
    lint_args, lint_ctx = setup_lint(ctx, lint_args=lint_args, **kwds)
    lint_ctx.lint(
        "lint_expansion",
        lint_expansion,
        realized_repository,
    )
    lint_ctx.lint(
        "lint_expected_files",
        lint_expected_files,
        realized_repository,
    )
    lint_ctx.lint(
        "lint_tool_dependencies_xsd",
        lint_tool_dependencies_xsd,
        realized_repository,
    )
    lint_ctx.lint(
        "lint_tool_dependencies_sha256sum",
        lint_tool_dependencies_sha256sum,
        realized_repository,
    )
    lint_ctx.lint(
        "lint_tool_dependencies_actions",
        lint_tool_dependencies_actions,
        realized_repository,
    )
    lint_ctx.lint(
        "lint_repository_dependencies",
        lint_repository_dependencies,
        realized_repository,
    )
    lint_ctx.lint(
        "lint_shed_yaml",
        lint_shed_yaml,
        realized_repository,
    )
    lint_ctx.lint(
        "lint_readme",
        lint_readme,
        realized_repository,
    )
    if kwds["urls"]:
        lint_ctx.lint(
            "lint_urls",
            lint_tool_dependencies_urls,
            realized_repository,
        )
    if kwds["tools"]:
        tools_failed = lint_repository_tools(ctx, realized_repository, lint_ctx, lint_args)
        failed = failed or tools_failed
    if kwds["ensure_metadata"]:
        lint_ctx.lint(
            "lint_shed_metadata",
            lint_shed_metadata,
            realized_repository,
        )
    return handle_lint_complete(lint_ctx, lint_args, failed=failed)


def lint_repository_tools(ctx, realized_repository, lint_ctx, lint_args):
    path = realized_repository.path
    for (tool_path, tool_source) in yield_tool_sources(ctx, path, recursive=True):
        original_path = tool_path.replace(path, realized_repository.real_path)
        info("+Linting tool %s" % original_path)
        if handle_tool_load_error(tool_path, tool_source):
            return True
        lint_tool_source_with(lint_ctx, tool_source, extra_modules=lint_args["extra_modules"])


def lint_expansion(realized_repository, lint_ctx):
    missing = realized_repository.missing
    if missing:
        msg = "Failed to expand inclusions %s" % missing
        lint_ctx.warn(msg)
    else:
        lint_ctx.info("Included files all found.")


def lint_shed_metadata(realized_repository, lint_ctx):
    found_all = True
    for key in SHED_METADATA:
        if key not in realized_repository.config:
            found_all = False
            lint_ctx.warn("Missing shed metadata field [%s] for repository" % key)
    if found_all:
        lint_ctx.info("Found all shed metadata fields required for automated repository " "creation and/or updates.")


def lint_readme(realized_repository, lint_ctx):
    path = realized_repository.real_path
    readme_rst = os.path.join(path, "README.rst")
    readme = os.path.join(path, "README")
    readme_txt = os.path.join(path, "README.txt")

    readme_found = False
    for readme in [readme_rst, readme, readme_txt]:
        if os.path.exists(readme):
            readme_found = readme

    readme_md = os.path.join(path, "README.md")
    if not readme_found and os.path.exists(readme_md):
        lint_ctx.warn("Tool Shed doesn't render markdown, " "README.md is invalid readme.")
        return

    if not readme_found:
        # TODO: filter on TYPE and make this a warning if
        # unrestricted repository - need to update iuc standards
        # first though.
        lint_ctx.info("No README found skipping.")
        return

    if readme_found.endswith(".rst"):
        with open(readme_found) as fh:
            readme_text = fh.read()
        invalid_rst = rst_invalid(readme_text)
        if invalid_rst:
            template = "Invalid restructured text found in README [%s]."
            msg = template % invalid_rst
            lint_ctx.warn(msg)
            return
        lint_ctx.info("README found containing valid reStructuredText.")
    else:
        lint_ctx.info("README found containing plain text.")


def lint_tool_dependencies_urls(realized_repository, lint_ctx):
    path = realized_repository.real_path
    tool_dependencies = os.path.join(path, "tool_dependencies.xml")
    if not os.path.exists(tool_dependencies):
        lint_ctx.info("No tool_dependencies.xml, skipping.")
        return

    root = ET.parse(tool_dependencies).getroot()
    lint_urls(root, lint_ctx)


def lint_tool_dependencies_sha256sum(realized_repository, lint_ctx):
    tool_dependencies = os.path.join(realized_repository.real_path, "tool_dependencies.xml")
    if not os.path.exists(tool_dependencies):
        lint_ctx.info("No tool_dependencies.xml, skipping.")
        return

    root = ET.parse(tool_dependencies).getroot()

    count = 0
    for action in root.findall(".//action"):
        assert action.tag == "action"
        if action.attrib.get("type", "") not in ["download_by_url", "download_file"]:
            continue
        url = action.text.strip()
        checksum = action.attrib.get("sha256sum", "")
        if not checksum:
            lint_ctx.warn("Missing checksum for %s" % url)
        elif len(checksum) != 64 or not set("0123456789abcdef").issuperset(checksum.lower()):
            lint_ctx.error(f"Invalid checksum {checksum!r} for {url}")
        else:
            # TODO - See planned --verify option to check it matches
            # lint_ctx.info("SHA256 checkum listed for %s" % url)
            count += 1
    if count:
        lint_ctx.info("Found %i download action(s) with SHA256 checksums" % count)


def lint_tool_dependencies_xsd(realized_repository, lint_ctx):
    path = realized_repository.real_path
    tool_dependencies = os.path.join(path, "tool_dependencies.xml")
    if not os.path.exists(tool_dependencies):
        lint_ctx.info("No tool_dependencies.xml, skipping.")
        return
    lint_xsd(lint_ctx, TOOL_DEPENDENCIES_XSD, tool_dependencies)


def lint_tool_dependencies_actions(realized_repository, lint_ctx):
    path = realized_repository.real_path
    tool_dependencies = os.path.join(path, "tool_dependencies.xml")
    if not os.path.exists(tool_dependencies):
        lint_ctx.info("No tool_dependencies.xml, skipping.")
        return
    try:
        base.Dependencies(tool_dependencies)
        lint_ctx.info("Parsed tool dependencies.")
    except Exception as e:
        import sys
        import traceback

        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        traceback.print_exc()
        template = "Problem parsing tool_dependenies.xml [%s]"
        msg = template % unicodify(e)
        lint_ctx.warn(msg)
        return


def lint_expected_files(realized_repository, lint_ctx):
    if realized_repository.is_package:
        if not os.path.exists(realized_repository.tool_dependencies_path):
            lint_ctx.warn("Package repository does not contain a " "tool_dependencies.xml file.")

    if realized_repository.is_suite:
        if not os.path.exists(realized_repository.repo_dependencies_path):
            lint_ctx.warn("Suite repository does not contain a " "repository_dependencies.xml file.")


def lint_repository_dependencies(realized_repository, lint_ctx):
    path = realized_repository.real_path
    repo_dependencies = os.path.join(path, "repository_dependencies.xml")
    if not os.path.exists(repo_dependencies):
        lint_ctx.info("No repository_dependencies.xml, skipping.")
        return
    lint_xsd(lint_ctx, REPO_DEPENDENCIES_XSD, repo_dependencies)


def lint_shed_yaml(realized_repository, lint_ctx):
    path = realized_repository.real_path
    shed_yaml = os.path.join(path, ".shed.yml")
    if not os.path.exists(shed_yaml):
        lint_ctx.info("No .shed.yml file found, skipping.")
        return
    try:
        with open(shed_yaml) as fh:
            yaml.safe_load(fh)
    except Exception as e:
        lint_ctx.warn("Failed to parse .shed.yml file [%s]" % unicodify(e))
        return
    lint_ctx.info(".shed.yml found and appears to be valid YAML.")
    _lint_shed_contents(lint_ctx, realized_repository)


def _lint_shed_contents(lint_ctx, realized_repository):
    config = realized_repository.config

    def _lint_if_present(key, func, *args):
        value = config.get(key, None)
        if value is not None:
            msg = func(value, *args)
            if msg:
                lint_ctx.warn(msg)

    _lint_if_present("owner", validate_repo_owner)
    _lint_if_present("name", validate_repo_name)
    _lint_if_present("type", _validate_repo_type, config["name"])
    _lint_if_present("categories", _validate_categories, realized_repository)


def _validate_repo_type(repo_type, name):
    if repo_type not in VALID_REPOSITORY_TYPES:
        return "Invalid repository type specified [%s]" % repo_type

    is_dep = repo_type == "tool_dependency_definition"
    is_suite = repo_type == "repository_suite_definition"
    if is_dep and not name.startswith("package_"):
        return "Tool dependency definition repositories should have names " "starting with package_"
    if is_suite and not name.startswith("suite_"):
        return "Repository suite definition repositories should have names " "starting with suite_"
    if name.startswith("package_") or name.startswith("suite_"):
        if repo_type == "unrestricted":
            return "Repository name indicated specialized repository type " "but repository is listed as unrestricted."


def _validate_categories(categories, realized_repository):
    msg = None
    if len(categories) == 0:
        msg = "Repository should specify one or more categories."
    else:
        for category in categories:
            unknown_categories = []
            if category not in CURRENT_CATEGORIES:
                unknown_categories.append(category)
            if unknown_categories:
                msg = "Categories [%s] unknown." % unknown_categories
        if realized_repository.is_package:
            if "Tool Dependency Packages" not in categories:
                msg = (
                    "Packages should be placed and should only be placed " "in the category 'Tool Dependency Packages'."
                )

    return msg


__all__ = ("lint_repository",)



"""Module contianing the :class:`GalaxyEngine` implementation of :class:`Engine`."""

import abc
import contextlib

from galaxy.tool_util.verify import interactor

from planemo import io
from planemo.galaxy.activity import (
    execute,
    execute_rerun,
)
from planemo.galaxy.config import external_galaxy_config
from planemo.galaxy.serve import serve_daemon
from planemo.runnable import (
    DelayedGalaxyToolTestCase,
    ExternalGalaxyToolTestCase,
    GALAXY_TOOLS_PREFIX,
    RunnableType,
)
from .interface import BaseEngine

INSTALLING_MESSAGE = "Installing repositories - this may take some time..."


class GalaxyEngine(BaseEngine, metaclass=abc.ABCMeta):
    """An :class:`Engine` implementation backed by a managed Galaxy.

    More information on Galaxy can be found at http://galaxyproject.org/.
    """

    handled_runnable_types = [
        RunnableType.cwl_tool,
        RunnableType.cwl_workflow,
        RunnableType.galaxy_workflow,
        RunnableType.galaxy_tool,
        RunnableType.galaxy_datamanager,
    ]

    def _run(self, runnables, job_paths):
        """Run job in Galaxy."""
        results = []
        for runnable, job_path in zip(runnables, job_paths):
            self._ctx.vlog(f"Serving artifact [{runnable}] with Galaxy.")
            with self.ensure_runnables_served([runnable]) as config:
                self._ctx.vlog(f"Running job path [{job_path}]")
                if self._ctx.verbose:
                    self._ctx.log(f"Running Galaxy with API configuration [{config.user_api_config}]")
                run_response = execute(self._ctx, config, runnable, job_path, **self._kwds)
                results.append(run_response)

        return results

    @abc.abstractmethod
    def ensure_runnables_served(self, runnables):
        """Use a context manager and describe Galaxy instance with runnables being served."""

    def _run_test_cases(self, test_cases, test_timeout):
        test_results = []
        file_based_test_cases = []
        embedded_test_cases = []
        # TODO: unify interface so we don't need to split test cases
        for test_case in test_cases:
            if isinstance(test_case, ExternalGalaxyToolTestCase):
                embedded_test_cases.append(test_case)
            else:
                file_based_test_cases.append(test_case)
        if file_based_test_cases:
            test_results.extend(super()._run_test_cases(file_based_test_cases, test_timeout))
        if embedded_test_cases:
            runnables = [test_case.runnable for test_case in embedded_test_cases]
            with self.ensure_runnables_served(runnables) as config:
                expanded_test_cases = expand_test_cases(config, embedded_test_cases)
                for test_case in expanded_test_cases:
                    galaxy_interactor_kwds = {
                        "galaxy_url": config.galaxy_url,
                        "master_api_key": config.master_api_key,
                        "api_key": config.user_api_key,
                        "keep_outputs_dir": self._kwds.get("test_data_target_dir"),
                    }
                    tool_id = test_case.tool_id
                    test_index = test_case.test_index
                    tool_version = test_case.tool_version
                    galaxy_interactor = interactor.GalaxyInteractorApi(**galaxy_interactor_kwds)

                    def _register_job_data(job_data):
                        test_results.append(
                            {
                                "id": tool_id + "-" + str(test_index),
                                "has_data": True,
                                "data": job_data,
                            }
                        )

                    verbose = self._ctx.verbose
                    try:
                        if verbose:
                            # TODO: this is pretty hacky, it'd be better to send a stream
                            # and capture the output information somehow.
                            interactor.VERBOSE_GALAXY_ERRORS = True

                        interactor.verify_tool(
                            tool_id,
                            galaxy_interactor,
                            test_index=test_index,
                            tool_version=tool_version,
                            register_job_data=_register_job_data,
                            maxseconds=test_timeout,
                            quiet=not verbose,
                        )
                    except Exception:
                        pass

        return test_results


class LocalManagedGalaxyEngine(GalaxyEngine):
    """An :class:`Engine` implementation backed by a managed Galaxy.

    More information on Galaxy can be found at http://galaxyproject.org/.
    """

    @contextlib.contextmanager
    def ensure_runnables_served(self, runnables):
        # TODO: define an interface for this - not everything in config would make sense for a
        # pre-existing Galaxy interface.
        with serve_daemon(self._ctx, runnables, **self._serve_kwds()) as config:
            if "install_args_list" in self._serve_kwds():
                self.shed_install(config)
            yield config

    def shed_install(self, config):
        kwds = self._serve_kwds()
        install_args_list = kwds["install_args_list"]
        install_deps = not kwds.get("skip_dependencies", False)
        print(INSTALLING_MESSAGE)
        io.info(INSTALLING_MESSAGE)
        for install_args in install_args_list:
            install_args["install_tool_dependencies"] = install_deps
            install_args["install_repository_dependencies"] = True
            install_args["new_tool_panel_section_label"] = "Shed Installs"
            config.install_repo(**install_args)
        try:
            config.wait_for_all_installed()
        except Exception:
            if self._ctx.verbose:
                print("Failed to install tool repositories, Galaxy log:")
                print(config.log_contents)
                print("Galaxy root:")
                io.shell(["ls", config.galaxy_root])
            raise

    def _serve_kwds(self):
        return self._kwds.copy()


class DockerizedManagedGalaxyEngine(LocalManagedGalaxyEngine):
    """An :class:`Engine` implementation backed by Galaxy running in Docker.

    More information on Galaxy can be found at http://galaxyproject.org/.
    """

    def _serve_kwds(self):
        serve_kwds = self._kwds.copy()
        serve_kwds["dockerize"] = True
        return serve_kwds


class ExternalGalaxyEngine(GalaxyEngine):
    """An :class:`Engine` implementation backed by an external Galaxy instance."""

    @contextlib.contextmanager
    def ensure_runnables_served(self, runnables):
        # TODO: ensure tools are available
        with external_galaxy_config(self._ctx, runnables, **self._kwds) as config:
            config.install_workflows()
            yield config

    def rerun(self, ctx, rerunnable, **kwds):
        with self.ensure_runnables_served([]) as config:
            rerun_response = execute_rerun(ctx, config, rerunnable, **kwds)
            return rerun_response


def expand_test_cases(config, test_cases):
    expanded_test_cases = []
    for test_case in test_cases:
        if not isinstance(test_case, DelayedGalaxyToolTestCase):
            expanded_test_cases.append(test_case)
        else:
            runnable = test_case.runnable
            tool_id = runnable.uri.split(GALAXY_TOOLS_PREFIX)[1]
            test_data = config.gi.tools._get(f"{tool_id}/test_data")
            for test_dict in test_data:
                expanded_test_cases.append(
                    ExternalGalaxyToolTestCase(
                        runnable,
                        tool_id=tool_id,
                        tool_version=test_dict["tool_version"],
                        test_index=test_dict["test_index"],
                        test_dict=test_dict,
                    )
                )
    return expanded_test_cases


__all__ = (
    "DockerizedManagedGalaxyEngine",
    "ExternalGalaxyEngine",
    "LocalManagedGalaxyEngine",
)



"""Module describing the planemo ``shed_create`` command."""
import sys

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function
from planemo.io import info


@click.command("shed_create")
@options.shed_publish_options()
@options.shed_message_option()
@options.shed_skip_upload()
@command_function
def cli(ctx, paths, **kwds):
    """Create a repository in a Galaxy Tool Shed.

    This will read the settings from the ``.shed.yml`` file.
    """
    shed_context = shed.get_shed_context(ctx, **kwds)

    def create(realized_repository):
        repo_id = realized_repository.find_repository_id(ctx, shed_context)
        if repo_id is None:
            if realized_repository.create(ctx, shed_context):
                info("Repository created")
                if not kwds["skip_upload"]:
                    return shed.upload_repository(ctx, realized_repository, **kwds)
                else:
                    return 0
            else:
                return 2
        else:
            return 1

    exit_code = shed.for_each_repository(ctx, create, paths, **kwds)
    sys.exit(exit_code)



"""Module contains code for the Requirement, Reference and some general functions for training."""

import collections

import oyaml as yaml


class Requirement:
    """Class to describe a training requirement."""

    def __init__(self, req_type="internal", topic_name="introduction", title=None, tutorials=None, link=None):
        """Init a Requirement instance."""
        self.type = req_type
        self.topic_name = topic_name
        self.tutorials = tutorials
        self.title = title
        self.link = link

    def init_from_dict(self, metadata):
        """Init from a dictionary generated by export_to_ordered_dict."""
        self.type = metadata["type"]
        if self.type == "internal":
            self.topic_name = metadata["topic_name"]
            if "tutorials" in metadata:
                self.tutorials = metadata["tutorials"]
        else:
            self.title = metadata["title"]
            if self.type == "external":
                self.link = metadata["link"]

    def export_to_ordered_dict(self):
        """Export the requirement into an ordered dictionary."""
        req = collections.OrderedDict()
        req["type"] = self.type
        if self.type == "internal":
            req["topic_name"] = self.topic_name
            if self.tutorials:
                req["tutorials"] = self.tutorials
        else:
            req["title"] = self.title
            if self.type == "external":
                req["link"] = self.link
        return req


def load_yaml(filepath):
    """Load the content of a YAML file to a dictionary."""
    with open(filepath) as m_file:
        content = yaml.safe_load(m_file)
    return content


def save_to_yaml(content, filepath):
    """Save a dictionary to a YAML file."""
    with open(filepath, "w") as stream:
        yaml.safe_dump(
            content,
            stream,
            indent=2,
            default_flow_style=False,
            default_style="",
            explicit_start=True,
            encoding="utf-8",
            allow_unicode=True,
        )



"""Module describing the planemo ``autoupdate`` command."""
import json
from typing import Callable

import click
import yaml
from galaxy.tool_util.parser.xml import XmlToolSource

from planemo import (
    autoupdate,
    options,
)
from planemo.cli import command_function
from planemo.config import planemo_option
from planemo.engine import (
    engine_context,
    is_galaxy_engine,
)
from planemo.engine.test import test_runnables
from planemo.exit_codes import (
    EXIT_CODE_GENERIC_FAILURE,
    EXIT_CODE_OK,
)
from planemo.io import (
    coalesce_return_codes,
    error,
    info,
    temp_directory,
)
from planemo.runnable import (
    for_paths,
    RunnableType,
)
from planemo.tools import (
    is_tool_load_error,
    yield_tool_sources_on_paths,
)


def dry_run_option() -> Callable:
    """Perform a dry run autoupdate without modifying the XML files"""
    return planemo_option(
        "--dry-run", is_flag=True, help="Perform a dry run autoupdate without modifying the XML files."
    )


def test_option() -> Callable:
    """Test updated XML files"""
    return planemo_option("--test", is_flag=True, help="Test updated XML files.")


def skiplist_option() -> Callable:
    """List of XML files to skip"""
    return planemo_option(
        "--skiplist",
        default=None,
        help="Skiplist file, containing a list of tools or workflows for which autoupdate should be skipped.",
    )


def skip_requirements_option() -> Callable:
    """List of requirements to skip"""
    return planemo_option(
        "--skip_requirements",
        default="python,r-base,perl",
        help="Comma-separated list of requirements which should be not be updated. Default is python,r-base,perl.",
    )


@click.command("autoupdate")  # noqa C901
@options.optional_tools_arg(multiple=True)
@dry_run_option()
@options.recursive_option()
@test_option()
@skiplist_option()
@skip_requirements_option()
@options.test_options()
@options.galaxy_target_options()
@options.galaxy_config_options()
@options.report_level_option()
@options.report_xunit()
@options.fail_level_option()
@options.galaxy_url_option()
@options.galaxy_user_key_option()
@command_function
def cli(ctx, paths, **kwds):  # noqa C901
    """Auto-update tool requirements by checking against Conda and updating if newer versions are available."""
    assert_tools = kwds.get("assert_tools", True)
    recursive = kwds.get("recursive", False)
    exit_codes = []
    modified_files = set()
    tools_to_skip = [line.rstrip() for line in open(kwds["skiplist"])] if kwds["skiplist"] else []
    runnables = for_paths(paths)

    if any(r.type in {RunnableType.galaxy_tool, RunnableType.directory} for r in runnables):
        # update Galaxy tools
        for (tool_path, tool_xml) in yield_tool_sources_on_paths(ctx, paths, recursive):
            if tool_path.split("/")[-1] in tools_to_skip:
                info("Skipping tool %s" % tool_path)
                continue
            info("Auto-updating tool %s" % tool_path)
            try:
                updated = autoupdate.autoupdate_tool(ctx, tool_path, modified_files=modified_files, **kwds)
                if updated:
                    modified_files.update(updated)
            except Exception as e:
                error(f"{tool_path} could not be updated - the following error was raised: {e.__str__()}")
            if handle_tool_load_error(tool_path, tool_xml):
                exit_codes.append(EXIT_CODE_GENERIC_FAILURE)
                continue
            else:
                exit_codes.append(EXIT_CODE_OK)

    workflows = [
        r for r in runnables if r.type == RunnableType.galaxy_workflow and r.path.split("/")[-1] not in tools_to_skip
    ]
    modified_workflows = []
    for workflow in workflows:
        tools_to_update = autoupdate.get_tools_to_update(ctx, workflow, tools_to_skip)
        if tools_to_update:
            modified_workflows.append(workflow)
            info("The following tools are outdated:")
            for tool, versions in tools_to_update.items():
                info(f"{tool}: {versions['current']} -> {versions['updated']}")
        else:
            info("No newer tool versions were found, so the workflow was not updated.")

    if modified_workflows and not kwds.get("dry_run"):
        assert is_galaxy_engine(**kwds)
        if kwds.get("engine") != "external_galaxy":
            kwds["install_most_recent_revision"] = True
            kwds["install_resolver_dependencies"] = False
            kwds["install_repository_dependencies"] = False
            kwds["shed_install"] = True

        with engine_context(ctx, **kwds) as galaxy_engine:
            with galaxy_engine.ensure_runnables_served(modified_workflows) as config:
                for workflow in modified_workflows:
                    if config.updated_repos.get(workflow.path) or kwds.get("engine") == "external_galaxy":
                        info("Auto-updating workflow %s" % workflow.path)
                        updated_workflow = autoupdate.autoupdate_wf(ctx, config, workflow)

                        if workflow.path.endswith(".ga"):
                            with open(workflow.path) as f:
                                original_workflow = json.load(f)
                            edited_workflow = autoupdate.fix_workflow_ga(original_workflow, updated_workflow)
                            with open(workflow.path, "w") as f:
                                json.dump(edited_workflow, f, indent=4)
                        else:
                            with open(workflow.path) as f:
                                original_workflow = yaml.load(f, Loader=yaml.SafeLoader)
                            edited_workflow = autoupdate.fix_workflow_gxformat2(original_workflow, updated_workflow)
                            with open(workflow.path, "w") as f:
                                yaml.dump(edited_workflow, f)
                        if original_workflow.get("release"):
                            info(
                                f"The workflow release number has been updated from "
                                f"{original_workflow.get('release')} to {edited_workflow.get('release')}."
                            )

    if kwds["test"]:
        if not modified_files and not modified_workflows:
            info("No tools or workflows were updated, so no tests were run.")
        else:
            with temp_directory(dir=ctx.planemo_directory) as temp_path:
                # only test tools in updated directories
                modified_paths = [
                    path
                    for path, tool_xml in yield_tool_sources_on_paths(ctx, paths, recursive)
                    if path in modified_files
                ]
                info(f"Running tests for the following auto-updated tools: {', '.join(modified_paths)}")
                runnables = for_paths(modified_paths + modified_workflows, temp_path=temp_path)
                kwds["engine"] = "galaxy"
                return_value = test_runnables(ctx, runnables, original_paths=paths, **kwds)
                exit_codes.append(return_value)
    return coalesce_return_codes(exit_codes, assert_at_least_one=assert_tools)


def handle_tool_load_error(tool_path: str, tool_xml: XmlToolSource) -> bool:
    """Return True if tool_xml is tool load error (invalid XML), and
    print a helpful error message.
    """
    is_error = False
    if is_tool_load_error(tool_xml):
        info("Could not update %s due to malformed xml." % tool_path)
        is_error = True
    return is_error
